from __future__ import annotations

import gzip
from collections import defaultdict
from math import fabs
from pathlib import Path
from typing import Any
from typing import Callable
from typing import cast
from typing import IO
from typing import Iterator
from typing import Literal
from typing import TypeAlias
from typing import TypedDict

import numpy as np
import pandas as pd
from lxml import etree  # type: ignore

from .logger import logger
from .unimod import Unimod
from .unimod import UnimodLookup

Converters: TypeAlias = dict[str, Callable[[Any], Any]]
FP_UNDERFLOW = np.finfo(np.float32).tiny

# see http://sashimi.sourceforge.net/schema_revision/pepXML/pepXML_v122.xsd

pi = "http://regis-web.systemsbiology.net/pepXML"

# TAGS...
Search_summary = "{%s}search_summary" % pi
Spectrum_query = "{%s}spectrum_query" % pi
Roc_error_data = "{%s}roc_error_data" % pi
File = "File"
MissingPP = "MissingPP"

NSMAP = {"p": pi}


def maybefloat(v: str | None) -> float | None:
    return v if v is None else float(v)


def maybeint(v: str | None) -> int | None:
    return v if v is None else int(v)


def Id(v: Any) -> Any:
    assert v is not None
    return v


def ynToBool(v: str) -> bool:
    return v.upper() == "Y"


def maybeoneZeroToBool(v: str | None) -> None | bool:
    return v if v is None else v == "1"


def tod(D: Converters, node: etree.Element, force: bool = False) -> dict[str, Any]:
    a = node.attrib
    if force:
        return {k: D[k](a[k]) for k in D}
    else:
        return {k: D[k](a[k]) for k in D if k in a}


def name_v(i: list[etree.Element]) -> dict[str, str]:
    return {n.attrib["name"]: n.attrib["value"] for n in i}


def prevent_underflow(v: Any) -> float:
    """These values go to FLOAT columns. Ensure no underflow"""
    v = float(v)
    return 0.0 if fabs(v) < FP_UNDERFLOW else v


COMET = {
    "xcorr": "score1",
    "deltacn": "score2",
    "deltacnstar": "score3",
    "spscore": "score4",
    "sprank": "score5",
    "expect": "score6",
}

MASCOT = {
    "ionscore": "score1",
    "identityscore": "score2",
    "star": "score3",
    "homologyscore": "score4",
    "expect": "score5",  # agh!
}


def chop_spectrum(spectrum: str) -> str:
    # return spectrum.rsplit('.', maxsplit=3)[0]
    ret = ".".join(spectrum.split(".")[:-3])
    return ret


def mods2json(mods: list[dict]) -> list[dict]:
    if not mods:
        return []

    def ml() -> Iterator[dict]:
        for m in mods:
            if "unimod" in m:
                u = m["unimod"]

                yield dict(**m, unimodId=u.unimod_id, mass=u.mono_delta)
            yield m

    return sorted(ml(), key=lambda d: d["position"])


def rowgetters(
    items: dict[str, Any],
    paranoid: bool = False,
) -> Iterator[tuple[str, Callable]]:  # noqa C901:
    def mkget(keys: list[str], check: bool) -> Callable[[Any], Any]:
        *rel, attr = keys

        def func(o: Any) -> Any:
            if check and not o:
                return None
            for k in rel:
                o = o[k]
                if check and not o:
                    return None
            if paranoid and o is None:
                raise RuntimeError(f"none type access for keys {keys} attr {attr}")
            if attr in o:
                return o[attr]
            return None

        return func

    def mkgetm(
        keys: list[list[str]],
        check: bool,
    ) -> Callable[[Any], Any]:  # pragma: no cover
        attr = []
        rel = []
        assert len({len(k) for k in keys}) == 1, keys
        for kl in keys:
            *rl, a = kl
            attr.append(a)
            rel.append(rl)

        def func(o: Any) -> None | list[Any]:
            if check and not o:
                return None
            ol = [o] * len(keys)
            for kv in zip(*rel):
                ol = [o[k] for o, k in zip(o, kv)]
                if check and not all(ol):
                    return None

            return [o[a] for o, a in zip(ol, attr)]

        return func

    def mkfunc(getter: Callable, f: Callable) -> Callable:
        def func(o: Any) -> Any:
            return f(getter(o))

        return func

    def mkfuncm(getter: Callable, f: Callable) -> Callable:
        def func(o: Any) -> Any:
            return f(*getter(o))

        return func

    def noop(val: Any) -> Any:
        return val

    for col, v in items.items():
        if isinstance(v, tuple):
            if len(v) == 3:
                key, cvt, check = v
            else:
                check = False
                key, cvt = v
                if isinstance(cvt, bool):
                    check = cvt
                    cvt = noop
        else:
            key, cvt, check = v, noop, False
        if isinstance(key, tuple):
            yield col, mkfuncm(mkgetm([k.split(",") for k in key], check=check), cvt)
        else:
            yield col, mkfunc(mkget(key.split("."), check=check), cvt)


def modified_peptide(peptide: str, modifications: list[dict[str, Any]]) -> str:
    if not modifications:
        return peptide
    pos = np.array([d["position"] for d in modifications], dtype=np.int32)

    end = len(peptide) + 1
    if pos[0] == 0:
        pos = pos[1:]
    if len(pos) > 0 and pos[-1] == end:
        pos = pos[:-1]
    if len(pos) == 0:
        return peptide
    frags = ["".join(p) for p in np.split(list(peptide), pos)]
    return (
        "".join(f"{f}[{int(m['mass'])}]" for f, m in zip(frags[:-1], modifications))
        + frags[-1]
    )


def spectrum_to_pyteomics_row(  # noqa: C901
    *,
    decoy_prefix: str = "decoy_",
    paranoid: bool = False,
) -> Callable[[dict[str, Any]], Iterator[dict[str, Any]]]:
    # create a pyteomics.pepxml equivalent row
    decoy_prefix = decoy_prefix.lower()

    def is_decoy(accession: str, dprefix: str) -> bool:
        return accession.lower().startswith(dprefix)

    def absolute_delta(v: float) -> float:
        av = fabs(v)
        return 0.0 if av < FP_UNDERFLOW else av

    def first_peptide(modt: list[tuple[str, Any]]) -> str | None:
        for pep, _ in modt:
            return pep
        return None

    GET = dict(
        # run_id=("spectrum.spectrum", sample_mod.run_id),
        # sample_fk=("spectrum.spectrum", sample_mod.sample_fk),
        peptide="hit.peptide",
        # peptideprophet_probability=("hit.peptide_prophet.probability", not expect_pp),
        # interprophet_probability="hit.interprophet.probability",
        # fval=("hit.peptide_prophet.fval", not expect_pp),
        # start_scan_nr="spectrum.start_scan",
        # hit_rank="hit.hit_rank",
        precursor_neutral_mass="spectrum.precursor_neutral_mass",
        calc_neutral_pep_mass="hit.calc_neutral_pep_mass",
        absolute_delta=("hit.massdiff", absolute_delta),
        assumed_charge="spectrum.assumed_charge",
        retention_time_sec="spectrum.retention_time_sec",
        num_missed_cleavages="hit.num_missed_cleavages",
        modifications=(
            "hit.mods",
            lambda modt: mods2json([m for pep, mods in modt for m in mods]),
        ),
        modified_peptide=(
            "hit.mods",
            first_peptide,
        ),  # this is wrong sometimes it show no modifications... see code in pyteomics.pepxml
    )

    getters = dict(rowgetters(GET, paranoid=paranoid))

    # def mapscores(hit):
    #     ret = hit["scores"]
    #     if "xcorr" in ret:
    #         ret["search_engine"] = "Comet"
    #     elif 'ionscore' in ret:
    #         ret["search_engine"] = "Mascot"
    #     return ret

    def row(spectrum: dict[str, Any]) -> Iterator[dict[str, Any]]:
        hits = spectrum["search_hits"]
        for hit in hits:
            if "decoy_prefix" in hit and hit["decoy_prefix"] is not None:
                dprefix = hit["decoy_prefix"].lower()
            else:
                dprefix = decoy_prefix
            decoy_only = all(is_decoy(p["protein"], dprefix) for p in hit["proteins"])
            has_real = any(not is_decoy(p["protein"], dprefix) for p in hit["proteins"])

            o = dict(hit=hit, spectrum=spectrum)
            row = {col: get(o) for col, get in getters.items()}

            row = {**hit, **spectrum, **row}
            row["is_decoy"] = decoy_only
            row["has_real"] = has_real
            # row.update(mapscores(hit))

            row["run"] = chop_spectrum(spectrum["spectrum"])
            proteins = hit["proteins"]

            row["protein_descr"] = [
                d["protein_descr"] if "protein_descr" in d else "" for d in proteins
            ]
            for col in [
                "protein",
                "peptide_prev_aa",
                "peptide_next_aa",
            ]:
                row[col] = [d[col] if col in d else None for d in proteins]
            for col in ["num_tol_term"]:
                row[col] = [d[col] if col in d else -1 for d in proteins]
            if proteins and "num_tot_proteins" in proteins[0]:
                row["num_tot_proteins"] = proteins[0]["num_tot_proteins"]

            if "peptide_prophet" in row:
                pp = row.pop("peptide_prophet")
                if pp is not None:
                    if "all_ntt_prob" in pp:
                        pp["peptideprophet_ntt_prob"] = [
                            float(v) for v in pp.pop("all_ntt_prob")[1:-1].split(",")
                        ]
                    if "probability" in pp:
                        pp["peptideprophet_probability"] = pp.pop("probability")
                    row.update(pp)

            if "interprophet" in row:
                pp = row.pop("interprophet")
                if pp is not None:
                    if "all_ntt_prob" in pp:
                        pp["interprophet_ntt_prob"] = [
                            float(v) for v in pp.pop("all_ntt_prob")[1:-1].split(",")
                        ]
                    if "probability" in pp:
                        pp["interprophet_probability"] = pp.pop("probability")
                    row.update(pp)

            # sheesh!
            if "modified_peptide" not in row or row["modified_peptide"] is None:
                row["modified_peptide"] = row["peptide"]

            if "modifications" in row:
                m = row["modifications"]
                if m and row["peptide"] == row["modified_peptide"]:
                    row["modified_peptide"] = modified_peptide(
                        row["peptide"],
                        row["modifications"],
                    )
            # cleanup
            for col in [
                "scores",
                "mods",
                "isProteinTerm",
                "probability",
                "proteins",
                # "peptide_prophet",
                "search_hits",
                # "retention_time_sec",
                "all_ntt_prob",
            ]:
                row.pop(col, None)
            row["proteins"] = row.pop("protein")
            if "xcorr" in row:
                row["search_engine"] = "Comet"
                row["searchEngineScore"] = row["xcorr"]  # COMET
            elif "ionscore" in row:
                row["search_engine"] = "Mascot"
                row["searchEngineScore"] = row["ionscore"]  # MASCOT
            # elif "expect" in row:
            #     row["search_engine"] = "Unknown"
            #     row["searchEngineScore"] = row["expect"]
            yield row

    return row


def nodes(n: etree.Element) -> str:
    return etree.tostring(n).decode("utf-8")


# pylint: disable=too-many-nested-blocks
def add_unimod_to_mods(
    mods: list[tuple[Any, list[Any]]],
    unimod_lookup: UnimodLookup,
    peptide: str,
    isProteinTerm: Literal["N", "C"] | None,
) -> int:
    nmissing = 0

    def show(uu: list[tuple[Unimod, Any]]) -> str:
        return "{}".format(",".join([str(u.unimod_id) for u, s in uu]))

    for __mod_pep, mod in mods:
        for mm in mod:
            unimods = unimod_lookup.matchModification(
                peptide,
                mm["position"],
                mm.get("massdiff"),
                isProteinTerm,
            )
            if unimods:
                if len(unimods) > 1:
                    # what todo?
                    # iscommon = [t for t in unimods if t[1]]  # any common?
                    # one common match is OK
                    # if len(iscommon) > 1:  # more than one uncommon mod
                    #     click.secho(
                    #         f"multiple (common) unimod matches for {mm} {show(iscommon)}",
                    #         fg="blue",
                    #     )
                    # elif len(iscommon) == 0:
                    #     click.secho(
                    #         f"multiple (uncommon) unimod matches for {mm} {show(unimods)}",
                    #         fg="blue",
                    #     )
                    # else:
                    #     if verbose:
                    #         click.secho(
                    #             f"multiple unimod matches for {mm} {show(unimods)}",
                    #             fg="blue",
                    #         )
                    # choose common
                    unimods = sorted(unimods, key=lambda t: t[1], reverse=True)

                unimod, _iscommon_ = unimods[0]
                mm["unimod"] = unimod  # choose first uncommon
            else:
                # possibly fatal too..
                # click.secho(f"no unimod match for {mm}", fg="red")
                nmissing += 1
    return nmissing


def assertit(val: Any, node: Any) -> None:
    if not val:
        assert val, nodes(node)


class Mod(TypedDict):
    aminoacid: str
    mass: float
    massdiff: float
    variable: bool


class SearchSummary(TypedDict):
    search_engine: str
    mods: list[Mod]
    decoy_prefix: str | None
    mascotParams: list[str] | None


def peptide_maker(  # noqa C901:
    events: Iterator[tuple[Literal["start", "end"], etree.Element]],
    *,
    unimod_lookup: UnimodLookup | None = None,
    search_summary: SearchSummary | None = None,
    use_unimod: bool = True,
) -> Iterator[tuple[str, Any]]:  # noqa: C901
    n_pp_missing = 0

    if unimod_lookup is None:
        unimod_lookup = UnimodLookup([])

    def search_summary_func(n: etree.Element) -> SearchSummary:
        def mods() -> Iterator[Mod]:
            AA: Converters = {
                "aminoacid": Id,
                "mass": float,
                "massdiff": float,
                "variable": ynToBool,
                # 'symbol' : "*"
            }
            for aa in n.xpath(".//p:aminoacid_modification", namespaces=NSMAP):
                yield cast(Mod, tod(AA, aa, force=True))

        params = name_v(n.xpath(".//p:parameter", namespaces=NSMAP))
        decoy_prefix = params.get("decoy_prefix")
        mascotParams_str = params.get("IT_MODS")
        if mascotParams_str is not None:
            mascotParams = [t.strip()[:-4] for t in mascotParams_str.strip().split(",")]
        else:
            mascotParams = None

        return SearchSummary(
            search_engine=n.attrib["search_engine"],  # required
            mods=list(mods()),
            decoy_prefix=decoy_prefix,
            mascotParams=mascotParams,
        )

    def unimodlookup() -> UnimodLookup:
        assert unimod_lookup is not None
        if search_summary is None:
            return unimod_lookup
        known_mods = search_summary["mascotParams"]
        return unimod_lookup.create(known_mods)

    def search_hit(node: etree.Element) -> dict[str, Any]:
        nonlocal n_pp_missing

        PROT: Converters = {
            "protein": Id,
            "protein_descr": lambda t: t,
            "peptide_prev_aa": Id,
            "peptide_next_aa": Id,
            "num_tol_term": int,
            "num_tot_proteins": int,
        }

        def protein(n: etree.Element, force: bool = False) -> dict[str, Any]:
            return tod(PROT, n, force=force)

        proteins = [
            protein(p) for p in node.xpath(".//p:alternative_protein", namespaces=NSMAP)
        ]

        proteins = [protein(node, force=False)] + proteins
        peptide: str = node.attrib["peptide"]
        # nump = proteins[0]["num_tot_proteins"]
        # if not nump == len(proteins):
        #     logger.warning(
        #         "num_tot_proteins mismatch %s %d %d",
        #         peptide,
        #         nump,
        #         len(proteins),
        #     )

        _scores = name_v(node.xpath(".//p:search_score", namespaces=NSMAP))

        SC: dict[str, Callable[[Any], int | float]] = {
            "xcorr": prevent_underflow,
            "deltacn": prevent_underflow,
            "deltacnstar": prevent_underflow,
            "spscore": prevent_underflow,
            "sprank": int,
            "expect": prevent_underflow,
            "ionscore": prevent_underflow,
            "identityscore": prevent_underflow,
            "star": int,
            "homologyscore": prevent_underflow,
            "hyperscore": prevent_underflow,
            "nextscore": prevent_underflow,
        }

        scores = {k: SC[k](_scores[k]) for k in _scores if k in SC}

        def doPP(ret: dict[str, float | str], p: Any, cdict: Converters) -> None:
            for attr in ["probability", "all_ntt_prob"]:
                v = p.attrib[attr]
                ret[attr] = float(v) if attr == "probability" else v
            pp = name_v(
                p.xpath(".//p:search_score_summary/p:parameter", namespaces=NSMAP),
            )

            ret.update({k: cdict[k](pp[k]) for k in cdict if k in pp})

        PP: Converters = {
            "fval": float,
            "ntt": float,
            "nmc": float,
            "massd": float,
            "isomassd": float,
        }  # stored as fval
        pep_prophet: dict[str, float | str] = {}
        for p in node.xpath(
            ".//p:peptideprophet_result",
            namespaces=NSMAP,
        ):
            doPP(pep_prophet, p, PP)

        IP: Converters = {
            "nrs": float,
            "nsi": float,
            "nsm": float,
        }
        interprophet: dict[str, float | str] = {}
        for p in node.xpath(
            ".//p:interprophet_result",
            namespaces=NSMAP,
        ):
            doPP(interprophet, p, IP)

        def modification(
            peptide: str,
            m: etree.Element,
        ) -> tuple[str | None, dict[str, Any]]:
            M: Converters = {
                "mass": float,
                "position": int,
                "static": float,
                "variable": float,
            }
            ret = tod(M, m, force=False)
            if "variable" in ret:
                v = ret["variable"]
                if v is not None:
                    ret["massdiff"] = v
            elif "static" in ret:
                v = ret["static"]
                if v is not None:
                    ret["massdiff"] = v

            mass = ret["mass"]
            assert search_summary is not None
            md = [m["massdiff"] for m in search_summary["mods"] if m["mass"] == mass]
            assert md, (md, mass)
            ret["massdiff"] = md[0]
            pos = ret["position"]
            ret["site"] = peptide[pos - 1]
            p = m.getparent()
            return (
                (
                    p.attrib["modified_peptide"]
                    if "modified_peptide" in p.attrib
                    else None
                ),
                ret,
            )

        modsd: dict[str | None, list[dict[str, Any]]] = defaultdict(list)

        for pep, mod in [
            modification(peptide, m)
            for m in node.xpath(
                ".//p:modification_info/p:mod_aminoacid_mass",
                namespaces=NSMAP,
            )
        ]:
            modsd[pep].append(mod)

        for m in node.xpath(".//p:modification_info", namespaces=NSMAP):
            if "mod_nterm_mass" in m.attrib:
                modsd[m.attrib["modified_peptide"]].append(
                    {"position": 0, "mass": float(m.attrib["mod_nterm_mass"])},
                )
            if "mod_cterm_mass" in m.attrib:
                modsd[m.attrib["modified_peptide"]].append(
                    {
                        "position": 1 + len(peptide),
                        "mass": float(m.attrib["mod_nterm_mass"]),
                    },
                )
        mods = list(modsd.items())
        HIT: Converters = {
            "hit_rank": int,  # stored as hit_rank
            "num_matched_ions": int,
            "num_matched_peptides": int,
            "tot_num_ions": int,
            "calc_neutral_pep_mass": float,  # as theoretical mass ?
            "num_missed_cleavages": int,  # stored as missed_cleavages
            # "is_rejected": maybeoneZeroToBool,
            "massdiff": lambda v: (
                0.0 if v.startswith("+-") else float(v)
            ),  # as abs of absolute_delta
        }

        hit = tod(HIT, node, force=False)  # is_rejected missing

        assert not (
            {"massdiff", "hit_rank", "calc_neutral_pep_mass", "num_missed_cleavages"}
            - set(hit)
        ), (
            hit,
            nodes(node),
        )
        assert search_summary is not None
        prefix = search_summary["decoy_prefix"]

        decoy_prefix = prefix.lower() if prefix else None

        def is_a_decoy(a: str) -> bool:
            if decoy_prefix is None:
                return False
            return a.lower().startswith(decoy_prefix)

        # not trustworthy
        if prefix:
            is_decoy = all(is_a_decoy(p["protein"]) for p in proteins)
        else:
            is_decoy = False

        isProteinTerm: Literal["N", "C"] | None = (
            "N"
            if node.attrib["peptide_prev_aa"] == "-"
            else ("C" if node.attrib["peptide_next_aa"] == "-" else None)
        )

        if mods and use_unimod:
            add_unimod_to_mods(mods, unimodlookup(), peptide, isProteinTerm)

        hit["proteins"] = proteins
        hit["peptide"] = peptide  # stored as sequence
        hit["scores"] = scores
        hit["mods"] = mods
        hit["is_decoy"] = is_decoy
        hit["isProteinTerm"] = isProteinTerm
        hit["peptide_prophet"] = pep_prophet
        hit["interprophet"] = interprophet
        hit["decoy_prefix"] = decoy_prefix
        return hit

    def spectrum(n: etree.Element) -> dict[str, Any]:
        assert search_summary is not None

        hits = [search_hit(s) for s in n.xpath(".//p:search_hit", namespaces=NSMAP)]

        SPECTRUM: Converters = {
            "spectrum": Id,
            "start_scan": int,
            "end_scan": int,
            "precursor_neutral_mass": float,
            "uncalibrated_precursor_neutral_mass": float,
            "assumed_charge": int,  # as charge
            "index": int,
            "retention_time_sec": float,  # as retention_time * 1e-9 i.e. nano seconds
        }

        ret = tod(SPECTRUM, n, force=False)
        ret["search_hits"] = hits
        return ret

    def clear(n: etree.Element) -> None:
        n.clear()
        while n.getprevious() is not None:
            del n.getparent()[0]

    for action, n in events:
        if n.tag == Search_summary and action == "end":
            search_summary = search_summary_func(n)
            yield Search_summary, search_summary
            clear(n)
        elif n.tag == Spectrum_query and action == "end":
            assert search_summary is not None, n.tag
            yield Spectrum_query, spectrum(n)
            clear(n)
    if n_pp_missing:
        yield MissingPP, n_pp_missing


def iter_pepxml_fp(
    fp: IO[bytes],
    *,
    unimod_lookup: UnimodLookup | None = None,
) -> Iterator[tuple[str, Any]]:
    search_summary = None
    # pos = 0
    events = etree.iterparse(fp, events=("end",), huge_tree=True)

    for tag, pg in peptide_maker(
        events,
        unimod_lookup=unimod_lookup,
        search_summary=search_summary,
    ):
        if tag == Search_summary:
            if search_summary is not None:
                if search_summary != pg:
                    logger.warning("pepxml: different search_summary!")
            search_summary = pg
            # print('search summary', search_summary)
        else:
            yield tag, pg


def parse_pepxml(  # pragma: no cover
    pep_xmlfile: Path,
    *,
    unimod_lookup: UnimodLookup | None = None,
    use_unimod: bool = False,
) -> Iterator[tuple[str, Any]]:
    # parser = etree.XMLPullParser(("end",), huge_tree=True)
    # events = parser.read_events()

    def iterfp(
        fp: IO[bytes] | gzip.GzipFile,
        fp_raw: IO[bytes],
    ) -> Iterator[tuple[str, Any]]:
        search_summary = None
        pos = 0
        events = etree.iterparse(fp, events=("end",), huge_tree=True)

        for tag, pg in peptide_maker(
            events,
            unimod_lookup=unimod_lookup,
            search_summary=search_summary,
            use_unimod=use_unimod,
        ):
            if tag == Search_summary:
                if search_summary is not None:
                    if search_summary != pg:
                        logger.warning("different search_summary!")
                search_summary = pg
                # print('search summary', search_summary)
            else:
                yield tag, pg
            cpos = fp_raw.tell()
            yield File, cpos - pos
            pos = cpos
        cpos = fp_raw.tell()
        yield File, cpos - pos
        pos = cpos

    if pep_xmlfile.name.endswith(".gz"):
        with open(pep_xmlfile, "rb") as fpraw, gzip.open(fpraw, "rb") as fp:
            yield from iterfp(fp, fpraw)
    else:
        with open(pep_xmlfile, "rb") as fpraw:
            yield from iterfp(fpraw, fpraw)
    # parser.close()


def pepxml_dataframe(
    pepxml: Path,
    level: int = 0,
    number_of_bg_processes: int = 0,
    with_logger: bool = True,
) -> pd.DataFrame:
    from .config import PEPXML_CHUNKS
    from more_itertools import ichunked
    from .logger import log_iterator
    from .pepxml import scan_spectra

    rowf = spectrum_to_pyteomics_row(paranoid=True)

    def gen() -> Iterator[dict[str, Any]]:
        with open(pepxml, "rb") as fp:
            for tag, d in iter_pepxml_fp(fp):
                if tag == Spectrum_query:
                    row = rowf(d)

                    for d in row:
                        yield d  # first search_hit only
                        break

    if with_logger:
        total = sum(1 for _ in scan_spectra(pepxml))
        it = log_iterator(
            gen(),
            total=total,
            level=level,
            desc=f"{pepxml.name}",
            number_of_bg_processes=number_of_bg_processes,
        )
    else:
        it = gen()
    ret = []
    for chunk in ichunked(it, PEPXML_CHUNKS):
        df = pd.DataFrame(chunk)

        for col in df.columns:
            s = df[col]
            if (
                s.dtype.kind == "f"
                and s.dtype == np.float64
                and col
                not in {
                    "uncalibrated_precursor_neutral_mass",
                    "retention_time_sec",
                    "calc_neutral_pep_mass",
                    "precursor_neutral_mass",
                    "massdiff",
                }
            ):
                df[col] = s.astype(np.float32)
            elif s.dtype == "i" and s.dtype == np.int64 and col != "index":
                df[col] = s.astype(np.int32)
        ret.append(df)
    return pd.concat(ret, axis=0, ignore_index=True)


def okdf(df1: pd.DataFrame, df2: pd.DataFrame) -> None:  # pragma: no cover
    # modified_peptide is wrong...
    df2 = df2.set_index("spectrum")
    df1 = df1.set_index("spectrum")
    A = [
        "protein_descr",
        "modifications",
        "modified_peptide",
        "proteins",
        "num_tol_term",
        "peptide_prev_aa",
        "peptide_next_aa",
        "peptideprophet_ntt_prob",
    ]

    cols = list(set(df1.columns) & set(df2.columns))

    def okarr(s: pd.Series, c: str) -> bool:
        a1 = s[c]
        a2 = s[c + "_o"]
        if len(a1) != len(a2):
            return False
        return all(aa1 == aa2 for aa1, aa2 in zip(a1, a2))

    for c in cols:
        if c in A:
            continue
        v = (df1[c] == df2[c]).all()
        if not v:  # NaNs!
            if c == "is_decoy":
                print(c, False)
                continue
            nna = df1[c].isna().sum() + df2[c].isna().sum()
            if nna > 0:
                if np.nanmax(np.abs(df1[c] - df2[c])) != 0:
                    print(c, v)
            else:
                print(c, False)
    # check arrays
    for c in set(A) - {"modifications"}:
        j = df1[[c]].join(df2[[c]], rsuffix="_o")
        a = j.apply(lambda s: okarr(s, c), axis=1).all()
        if not a:
            print(c, a)

    # check modifications only insofar as mass and position are the same
    def fixmod(dl: list[dict[str, Any]]) -> str:
        n = 6
        return ":".join(f"{d['mass']:.{n}f}@{d['position']}" for d in dl)

    c = "modifications"
    d1 = df1[c].apply(fixmod)
    d2 = df2[c].apply(fixmod)
    if not all(d1 == d2):
        print(c, False)


class ErrorData(TypedDict):
    min_prob: float
    sensitivity: float
    error: float
    num_corr: int
    num_incorr: int


def error_data(node: etree.Element) -> list[ErrorData]:  # pragma: no cover
    ret: list[ErrorData] = []
    for e in node.xpath(".//p:roc_data_point", namespaces=NSMAP):
        args = {
            str(k): int(v) if k.startswith("num_") else float(v)
            for k, v in e.attrib.items()
        }
        ret.append(args)  # type: ignore

    return ret


def iter_roc_fp(  # pragma: no cover
    fp: IO,
) -> list[ErrorData]:
    def clear(n: etree.Element) -> None:
        n.clear()
        while n.getprevious() is not None:
            del n.getparent()[0]

    events = etree.iterparse(fp, events=("end",), huge_tree=True)
    for action, n in events:
        if n.tag == Spectrum_query and action == "end":
            clear(n)
            return []

        elif n.tag == Roc_error_data and action == "end":
            if n.attrib["charge"] == "all":
                return error_data(n)
            clear(n)
    return []


def error_data_dataframe(  # pragma: no cover
    pepxml: Path,
) -> pd.DataFrame:
    with pepxml.open("rb") as fp:
        return pd.DataFrame(iter_roc_fp(fp))



from __future__ import annotations

import logging
import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:
    from typing import Iterator


def hashandlers(logger: logging.Logger) -> bool:
    if os.environ.get("PYTEST_CURRENT_TEST") is not None:
        return bool(logger.handlers)
    return logger.hasHandlers()  # pragma: no cover


Result = TypeVar("Result")

logger = logging.getLogger("protein_turnover")


def init_logger(
    *,
    level: str | int = logging.INFO,
    logfile: Path | str | None = None,
    reinit: bool = False,
) -> None:
    from .config import LOG_FORMAT

    h: logging.Handler

    if isinstance(level, str):
        level = level.upper()

    def add_handler(h: logging.Handler) -> None:  # pragma: no cover
        fmt = logging.Formatter(LOG_FORMAT)
        h.setFormatter(fmt)
        logger.addHandler(h)

    logger.setLevel(level)
    if reinit:
        for h in list(logger.handlers):  # pragma: no cover
            logger.removeHandler(h)

    if not hashandlers(logger):
        if logfile is None or logfile == "-":
            add_handler(logging.StreamHandler())
        else:
            lf = Path(logfile)
            if not lf.parent.exists():
                lf.parent.mkdir(parents=True, exist_ok=True)
            add_handler(logging.FileHandler(str(logfile), mode="a"))


def log_iterator(
    it: Iterator[Result],
    *,
    total: int,
    desc: str,
    level: int = 1,
    number_of_bg_processes: int = 1,
    num_logs: int = 100,
) -> Iterator[Result]:
    n = max(1, total // num_logs)
    if number_of_bg_processes == 1:
        fmt = f"{desc}: [%d/%d %d%%]"
    else:
        fmt = f"{desc}[{level}/{number_of_bg_processes}]: [%d/%d %d%%]"

    for idx, i in enumerate(it, start=1):
        yield i
        if (idx % n) == 0:
            pct = int(idx * 100 / total)
            logger.info(fmt, idx, total, pct)
    logger.info(fmt, total, total, 100)


@contextmanager
def show_log():
    oldlevel = logger.getEffectiveLevel()
    h = None
    if not hashandlers(logger):  # local handler
        h = logging.StreamHandler()
        logger.addHandler(h)
    try:
        logger.setLevel(logging.INFO)
        yield
    finally:
        logger.setLevel(oldlevel)
        if h is not None:
            logger.removeHandler(h)


@contextmanager
def show_fit_warnings(quiet: bool = False):
    from . import fitenvelopes

    old = fitenvelopes.LOG_ERRORS
    oldlevel = None
    h = None
    if not quiet:
        fitenvelopes.LOG_ERRORS = True
        if not logger.isEnabledFor(logging.WARNING):  # pragma: no cover
            oldlevel = logger.getEffectiveLevel()
            logger.setLevel(logging.WARNING)
        if not logger.hasHandlers():  # pragma: no cover
            h = logging.StreamHandler()
            logger.addHandler(h)
    try:
        yield
    finally:
        if not quiet:
            fitenvelopes.LOG_ERRORS = old
            if oldlevel is not None:  # pragma: no cover
                logger.setLevel(oldlevel)
            if h is not None:  # pragma: no cover
                logger.removeHandler(h)



from __future__ import annotations

from typing import TypeAlias
from typing import TypeVar

import numpy as np

Number = TypeVar("Number", bound=np.number)

NP1DF64Array: TypeAlias = np.ndarray[tuple[int], np.dtype[np.float64]]
NP2DF64Array: TypeAlias = np.ndarray[tuple[int, int], np.dtype[np.float64]]
NP1DF32Array: TypeAlias = np.ndarray[tuple[int], np.dtype[np.float32]]
NP1DI32Array: TypeAlias = np.ndarray[tuple[int], np.dtype[np.int32]]
NP2DF32Array: TypeAlias = np.ndarray[tuple[int, int], np.dtype[np.float32]]
NP1DFArray: TypeAlias = np.ndarray[tuple[int], np.dtype[np.floating]]

NP3DF32Array: TypeAlias = np.ndarray[tuple[int, int, int], np.dtype[np.float32]]

NP1DNArray: TypeAlias = np.ndarray[tuple[int], np.dtype[Number]]



from __future__ import annotations

import re
import sys
import tomllib
import unicodedata
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import fields
from dataclasses import MISSING
from dataclasses import replace
from datetime import datetime
from os.path import commonprefix
from os.path import sep
from pathlib import Path
from typing import Any
from typing import Sequence

import tomli_w

from .resourcefiles import MzMLResourceFile
from .resourcefiles import PepXMLResourceFile
from .resourcefiles import ProtXMLResourceFile
from .resourcefiles import ResourceFiles
from .utils import PeptideSettings

try:
    from unidecode import unidecode  # type: ignore
except ImportError:  # pragma: no cover

    def unidecode(string: str, errors: str = "ignore", replace_str: str = "?") -> str:
        return unicodedata.normalize("NFKD", string)


WIDE = sys.maxunicode > 0xFFFF  # UCS-4 build of python


def find_prefix(filenames: Sequence[Path]) -> Path:
    prefix = commonprefix([Path(m).absolute() for m in filenames])
    if not prefix.endswith(sep):
        i = prefix.rfind(sep)
        if i > 0:
            prefix = prefix[: i + 1]
    return Path(prefix)


def jobidkey(
    pepxml: str,
    protxml: str,
    mzmlfiles: list[str],
) -> str:  # pragma: no cover
    from hashlib import md5

    m = md5()
    m.update(str(Path(pepxml).absolute()).encode("utf-8"))
    m.update(str(Path(protxml).absolute()).encode("utf-8"))
    for s in mzmlfiles:
        m.update(str(Path(s).absolute()).encode("utf-8"))

    return m.hexdigest()


## @export
def slugify(s: str, transliterate: bool = True) -> str:
    if not s:
        return s

    if WIDE and transliterate:  # UCS-4 build of python
        s = unidecode(s)
    else:  # pragma: no cover
        s = unicodedata.normalize("NFKD", s)

    slug = s.encode("ascii", "ignore").lower()
    slug = re.sub(b"[^a-z0-9]+", b"-", slug).strip(b"-")
    slug = re.sub(b"[-]+", b"-", slug)
    return slug.decode("ascii")


## @export
@dataclass(kw_only=True)
class TurnoverJob:
    job_name: str
    pepxml: list[str]
    mzmlfiles: list[str]
    protxml: str = ""
    jobid: str = ""
    settings: PeptideSettings = PeptideSettings()
    cache_dir: str | None = None
    email: str | None = None
    status: str | None = None  # pending, running, finished, failed, stopped, killed.
    mzfile_to_run: dict[str, str] | None = None
    match_runNames: bool = False
    """Match peptides to mzML files via their `spectrum` names"""
    aggregate_peptides: bool = True
    """Aggregate peptides and group by peptide, modification, assumed_charge and calc_neutral_pep_mass"""
    finished_time: datetime | None = None
    """Time when job was finished (only set by background runner)"""

    def __post_init__(self) -> None:
        if self.jobid == "":
            self.jobid = slugify(self.job_name)[:24]

    def cache_ok(self) -> bool:  # pragma: no cover
        return len(self.to_resource_files().todo()) == 0

    def get_mzfile_to_run(self) -> dict[str, str]:
        if self.mzfile_to_run is None:
            self.mzfile_to_run = {}
        for mz in self.mzmlfiles:
            pmz = Path(mz)
            if pmz.name not in self.mzfile_to_run:
                self.mzfile_to_run[pmz.name] = pmz.stem
        return self.mzfile_to_run

    @property
    def runNames(self) -> set[str]:
        return set(self.get_mzfile_to_run().values())

    def hash(self) -> str:
        from hashlib import md5

        m = md5()
        m.update(self.settings.hash().encode("utf-8"))
        for pepxml in self.pepxml:
            m.update(str(Path(pepxml).absolute()).encode("utf-8"))
        m.update(str(Path(self.protxml).absolute()).encode("utf-8"))
        for mzml in self.mzmlfiles:
            m.update(str(Path(mzml).absolute()).encode("utf-8"))
        for field in fields(self):
            if field.name in {
                "settings",
                "pepxml",
                "protxml",
                "mzmlfiles",
                "mzfile_to_run",
            }:
                continue
            v = getattr(self, field.name)
            m.update(str(v).encode("utf-8"))
        return m.hexdigest()

    def __hash__(self) -> int:
        return int(self.hash(), 16)

    def save_to_dir(
        self,
        directory: Path | str | None = None,
        filename: str | None = None,
    ) -> Path:
        outdir = Path(directory or self.cache_dir or ".")

        if filename is None:  # pragma: no cover
            out = outdir.joinpath(self.jobid + ".toml")
        else:
            if not filename.endswith(".toml"):
                filename += ".toml"
            out = outdir.joinpath(filename)

        if not out.parent.is_dir():  # pragma: no cover
            out.parent.mkdir(parents=True)

        return self.save_file(out)

    def save_file(self, out: Path) -> Path:
        mzmlfiles = [Path(m).absolute() for m in self.mzmlfiles]
        prefix = find_prefix(mzmlfiles)

        d = {k: v for k, v in asdict(self).items() if v is not None}

        if "cache_dir" in d:
            d["cache_dir"] = str(Path(d["cache_dir"]).expanduser().absolute())

        d["pepxml"] = [str(Path(m).absolute()) for m in self.pepxml]
        d["protxml"] = str(Path(d["protxml"]).absolute())
        d["mzmlfiles"] = [str(f.relative_to(prefix)) for f in mzmlfiles]
        d["mzmlprefix"] = str(Path(prefix).absolute())
        # d["settings"] = asdict(d["settings"])
        # settings = d.pop("settings")
        # d.update(settings)
        try:
            b = tomli_w.dumps(d).encode("utf-8")
            out.write_bytes(b)
        except Exception as e:  # pragma: no cover
            try:
                out.unlink(missing_ok=True)
            except OSError:
                pass
            raise e
        return out

    @classmethod
    def safe_jobid(
        cls,
        turnover_dict: dict[str, Any],
        filename: Path,
    ) -> str:  # pragma: no cover
        return filename.stem
        # job_name = str(turnover_dict["job_name"])
        # return safe_jobid(job_name, filename)

    @classmethod
    def restore(cls, filename: str | Path) -> TurnoverJob:
        def ensure_list(key: str) -> None:
            if key in turnover_dict:
                r = turnover_dict[key]
                if isinstance(r, str):
                    turnover_dict[key] = [r]

        filename = Path(filename).expanduser()
        with filename.open("rb") as fp:
            turnover_dict: dict[str, Any] = tomllib.load(fp)
            # just what we want...

            ensure_list("mzmlfiles")
            ensure_list("pepxml")
            if "job_name" not in turnover_dict:
                raise ValueError("please specify a job_name")
            if "jobid" not in turnover_dict:
                turnover_dict["jobid"] = cls.safe_jobid(turnover_dict, filename)
            if "mzmlprefix" in turnover_dict:
                prefix = Path(turnover_dict.pop("mzmlprefix"))
                turnover_dict["mzmlfiles"] = [
                    str(prefix.joinpath(m)) for m in turnover_dict["mzmlfiles"]
                ]
            if "settings" in turnover_dict:
                settings = turnover_dict["settings"]
                settingsd = {
                    f.name: settings[f.name]
                    for f in fields(PeptideSettings)
                    if f.name in settings
                }
                turnover_dict["settings"] = PeptideSettings(**settingsd)
            else:
                settingsd = {
                    f.name: turnover_dict[f.name]
                    for f in fields(PeptideSettings)
                    if f.name in turnover_dict
                }
                turnover_dict["settings"] = PeptideSettings(**settingsd)
            if "protxml" not in turnover_dict:
                turnover_dict["protxml"] = ""

            missing = {k for k in REQUIRED_FIELDS if k not in turnover_dict}
            if missing:
                s = "" if len(missing) > 1 else ""
                raise ValueError(
                    f'turnover file "{filename}" is missing: {", ".join(missing)} value{s}',
                )
            # cleanup
            turnover_dict = {k: v for k, v in turnover_dict.items() if k in ALL_FIELDS}
            return cls(**turnover_dict)

    def relative_to_path(self, path: Path | str) -> TurnoverJob:
        rep: dict[str, Any] = {}
        path = Path(path)

        def topath(p: str) -> str:
            return str(path.joinpath(Path(p)))

        if self.protxml != "":
            rep["protxml"] = topath(self.protxml)
        rep["pepxml"] = [topath(p) for p in self.pepxml]
        rep["mzmlfiles"] = [topath(p) for p in self.mzmlfiles]
        if self.cache_dir is not None:
            rep["cache_dir"] = topath(self.cache_dir)
        return replace(self, **rep)

    def verify(self) -> str | None:
        return verify_run(self)

    def to_resource_files(self) -> ResourceFiles:
        rf = ResourceFiles(
            [PepXMLResourceFile(m, self.cache_dir) for m in set(self.pepxml)],
            ProtXMLResourceFile(self.protxml, self.cache_dir),
            [MzMLResourceFile(m, self.cache_dir) for m in set(self.mzmlfiles)],
        )
        return rf


## @export
def remap_job(job: TurnoverJob, remapping: dict[str, str]) -> TurnoverJob:
    if not remapping:
        return job

    def remap(p: str) -> str:
        for k, v in remapping.items():
            if p.startswith(k):
                return v + p[len(k) :]
        return p  # pragma: no cover

    return replace(
        job,
        pepxml=[remap(f) for f in job.pepxml],
        protxml=remap(job.protxml),
        mzmlfiles=[remap(f) for f in job.mzmlfiles],
    )


ALL_FIELDS = {f.name for f in fields(TurnoverJob)}
REQUIRED_FIELDS = {f.name for f in fields(TurnoverJob) if f.default == MISSING}


def verify_run(
    job: TurnoverJob,
) -> str | None:
    # import pandas as pd
    # from .types.checking import check_pepxml_columns
    # from .utils import IO

    files = job.to_resource_files()
    msgs = []
    if not files.protxml.exists():
        msgs.append(f"no prot XML file: {files.protxml.original}")
    elif not files.protxml.cache_protxml_ok():
        msgs.append(f"cache for {files.protxml.original} is out of date or missing")

    # df = None
    for pepxml in files.pepxmls:
        if not pepxml.exists():
            msgs.append(f"no pep XML file: {pepxml.original}")
        elif not pepxml.cache_pepxml_ok():
            msg = f"cache for {pepxml.original.name} is out of date or missing"
            msgs.append(msg)
        # else:
        #     fname = pepxml.cache_pepxml()
        #     ndf = IO(fname).read_df()
        #     missing = check_pepxml_columns(ndf)
        #     if missing:
        #         msgs.append(f"columns missing {fname}: {missing}")

        #     if df is not None:
        #         df = pd.concat([df, ndf], axis=0)
        #     else:
        #         df = ndf

    # ok = True
    failed = [mzml for mzml in files.mzmlfiles if not mzml.exists()]
    if failed:
        for mzml in failed:
            msg = f"no mzML {mzml.original.name} file"
            msgs.append(msg)

        # ok = False

    failed = [
        mzml for mzml in files.mzmlfiles if mzml.exists() and not mzml.cache_mzml_ok()
    ]
    if failed:
        for mzml in failed:
            msg = f"cache for {mzml.original.name} is out of date or missing"
            msgs.append(msg)
        # ok = False

    # if ok and df is not None:
    #     if job.match_runNames:
    #         mzfile_to_run = job.get_mzfile_to_run()
    #         for mzml in files.mzmlfiles:
    #             hits = (df["run"] == mzfile_to_run[mzml.name]).sum()
    #             if hits == 0:
    #                 msg = f"no run hits in {mzml.original.name} for {mzml.name}"
    #                 msgs.append(msg)

    if not msgs:  # pragma: no cover
        return None
    return ", ".join(msgs)


def same_files(job1: TurnoverJob, job2: TurnoverJob) -> str | None:
    def issame(l1: list[str], l2: list[str]):
        p1 = [Path(p).absolute() for p in l1]
        p2 = [Path(p).absolute() for p in l2]
        if len(p1) != len(p2):
            return "different lengths"
        if not all(f1 == f2 for f1, f2 in zip(p1, p2)):
            return "files differ"
        return None

    msgs = []
    r = issame(job1.pepxml, job2.pepxml)
    if r is not None:
        msgs.append(f"pepxml: {r}")

    r = issame(job1.mzmlfiles, job2.mzmlfiles)
    if r is not None:
        msgs.append(f"mzml: {r}")

    if job1.protxml != "" or job2.protxml != "":
        r = issame([job1.protxml], [job2.protxml])
        if r is not None:
            msgs.append(f"protxml: {r}")
    if not msgs:
        return None
    return ", ".join(msgs)



from __future__ import annotations

import warnings
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from math import fabs
from typing import NamedTuple
from typing import TYPE_CHECKING
from typing import TypeVar

import numpy as np
import pandas as pd
from scipy.integrate import quad
from scipy.linalg import LinAlgError
from scipy.linalg import lstsq
from scipy.linalg import norm
from scipy.optimize import curve_fit
from scipy.optimize import nnls
from scipy.special import erf  # pylint: disable=no-name-in-module
from scipy.stats import pearsonr

from .algo import settings_to_algo
from .logger import logger
from .utils import ensure_pos
from .utils import PeptideInfo
from .utils import resize

if TYPE_CHECKING:
    from typing import Callable
    from typing import Iterator
    from .utils import PeptideSettings
    from .types.numpytypes import NP1DF32Array

FAIL_E_CURVE = 0
FAIL_W_CURVE = 1
FAIL_W_QUAD = 2
FAIL_W_H_CURVE = 3
FAIL_E_H_CURVE = 4
FAIL_W_PEARSON = 5
FAIL_E_BOUNDARY = 6
FAIL_E_EICS = 7
FAIL_W_PEARSON2 = 8
FAIL_E_NOEICS = 9

FAILS = {
    FAIL_E_CURVE: "envelope curve_fit error",
    FAIL_W_CURVE: "envelope curve_fit warning",
    FAIL_W_QUAD: "envelope quadrature warning",
    FAIL_W_H_CURVE: "heavy curve_fit warning",
    FAIL_E_H_CURVE: "heavy curve_fit error",
    FAIL_W_PEARSON: "pearson correlation warning for heavy vs theortical dist",
    FAIL_E_BOUNDARY: "no intensities within boundary",
    FAIL_E_EICS: "can't fit EICS",
    FAIL_W_PEARSON2: "pearson correlation warning for isotope vs theortical dist",
    FAIL_E_NOEICS: "no EICS found",
}

LOG_ERRORS = False  # see show_fit_warnings in logger.py

Result = TypeVar("Result")

R2: float = np.sqrt(2.0).astype(np.float64)
RPI2: float = np.sqrt(np.pi / 2.0).astype(np.float64)
NSIGMA = 2.0
DELTA_ERF: float = (RPI2 * (erf(NSIGMA / R2) - erf(-NSIGMA / R2))).astype(np.float64)
USE_QUAD = False
QUAD_LIMIT = 500


def fails(ival: int) -> str:
    return ", ".join(s for i, s in FAILS.items() if ival & (1 << i))


@dataclass
class IsotopeEnvelope:
    adjustedRsq: np.ndarray
    isotopeEnvelopes: np.ndarray  # 1-D
    monoFitParams: np.ndarray  # 1-D  [mu, sigma, scale, baseline]
    inv_ratio: float
    monoPeakArea: float
    maxPeakArea: float
    sigma: float
    avg_rsq: float


@dataclass
class LabelledEnvelope:
    labelledEnvelopes: np.ndarray  # 1-D
    theoreticalDist: np.ndarray
    relativeIsotopeAbundance: float
    enrichment: float
    # labelEnrichment2: float
    heavyCor: float
    heavyCor2: float
    nnls_residual: float
    totalNNLSWeight: float
    totalIntensityWeight: float


@dataclass
class Envelope(IsotopeEnvelope, LabelledEnvelope):
    @classmethod
    def make_all(
        cls,
        env: IsotopeEnvelope,
        # heavy: HeavyEnvelope,
        labelled: LabelledEnvelope,
    ) -> Envelope:
        args = {**asdict(env), **asdict(labelled)}
        return cls(**args)  # type: ignore


class RSquared(NamedTuple):
    r_squared: float
    adj_r_squared: float


def normalIntensityFunction(
    rt: float | np.ndarray,
    mu: float,
    sigma: float,
    scale: float,
    baseline: float,
) -> float | np.ndarray:
    return scale * np.exp(-0.5 * (rt - mu) ** 2 / sigma**2) + baseline


def lm_adjust(
    mat: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    df: int,
) -> RSquared:
    dx = mat @ x - y
    return adjust(dx, y, df)


# def curve_adjust(
#     func: Callable[[float | np.ndarray, float, float, float], float | np.ndarray],
#     popt: np.ndarray,
#     x: np.ndarray,
#     y: np.ndarray,
#     df: int,
# ) -> RSquared:
#     dx = func(x, *popt) - y
#     return adjust(dx, y, df)


# https://en.wikipedia.org/wiki/Coefficient_of_determination
def adjust(dx: np.ndarray, y: np.ndarray, df: int) -> RSquared:
    rss = np.sum(dx**2)
    ymean = np.mean(y)
    tss = np.sum((y - ymean) ** 2)
    if tss > 0.0:
        r_squared = 1.0 - rss / tss
        n = len(y)
        if n > df:
            adj_r_squared = 1.0 - ((n - 1) / (n - df)) * (1.0 - r_squared)
        else:
            adj_r_squared = r_squared
    else:
        r_squared, adj_r_squared = np.nan, np.nan
    return RSquared(r_squared, adj_r_squared)


def catch_warnings(
    func: Callable[[], Result],
    peptide: str,
    msg: str,
) -> tuple[Result, int]:  # Literal[0,1]
    with warnings.catch_warnings(record=True) as warn:
        warnings.simplefilter("always")
        ret = func()
    if warn and LOG_ERRORS:
        for w in warn:
            message = str(w.message)
            if "\n" in message:
                message, _ = message.split("\n", 1)

            logger.warning(msg, peptide, message)
    return ret, 1 if warn else 0


def heavy_dist(
    pepinfo: PeptideInfo,
    isotopeEnvelopesMinus: NP1DF32Array,  # without -1 isoDiff
    natural_dist: Callable[[PeptideInfo, float], NP1DF32Array],
) -> NP1DF32Array:
    """isotopeEnvelope - natural abundance"""
    naturalIsotopeEnvelope = natural_dist(pepinfo, isotopeEnvelopesMinus[0])
    naturalIsotopeEnvelope = resize(naturalIsotopeEnvelope, len(isotopeEnvelopesMinus))

    return ensure_pos(
        isotopeEnvelopesMinus - naturalIsotopeEnvelope,
    ).astype(np.float32)


def fitIntensities(
    mono_intensity: np.ndarray,  # float[N,2 or 1]
    intensities: np.ndarray,  # float[:,N]
) -> Iterator[list[float]]:
    with_origin = mono_intensity.shape[1] == 2
    for intensity in intensities:
        try:
            _res = lstsq(mono_intensity, intensity)
            assert _res is not None
            x, _residues, rank, _s = _res
            if with_origin:
                alpha, slope = x
            else:
                alpha, slope = 0.0, x[0]

            r = lm_adjust(mono_intensity, x, intensity, df=rank)
            adj_r_squared = r.adj_r_squared
        except LinAlgError as e:
            if LOG_ERRORS:
                logger.error("fitIntensities: %s", e)
            adj_r_squared, slope, alpha = np.nan, np.nan, np.nan
        yield [adj_r_squared, slope, alpha]  # type: ignore


@dataclass
class MonoFit:
    area: float
    lBoundary: float
    rBoundary: float
    fail: int = 0
    popt: np.ndarray = field(default_factory=lambda: np.zeros(4))
    pcov: np.ndarray = field(default_factory=lambda: np.zeros(16).reshape((4, 4)))

    @property
    def sigma(self) -> float:
        return float(np.average(np.sqrt(np.diag(self.pcov))))


def interpolateRT(
    rt: np.ndarray,
    intensity: np.ndarray,
    step: float = 0.5,
    *,
    rolling: int = 5,
) -> pd.DataFrame:
    n = len(rt)
    rtmin, rtmax = rt.min(), rt.max()
    if n > 0:
        s = (rtmax - rtmin) / n
        s = min(step, s)
    else:
        s = step
    interpolated_rt = np.arange(rtmin, rtmax + 1e-4, step=s)
    interpolated = lambda rtv: np.interp(rtv, rt, intensity)
    interpolated_i = interpolated(interpolated_rt)
    monoData = pd.DataFrame(dict(rt=interpolated_rt, int=interpolated_i))
    if rolling > 0:
        monoData["int"] = (
            monoData["int"].rolling(rolling, min_periods=1, center=True).mean()
        )

    return monoData


def fitMonoEnvelope(
    rt: np.ndarray,
    intensity: np.ndarray,
    interpolate: bool = True,
    use_quadrature: bool = False,
    peptide: str = "unk peptide",
) -> MonoFit:
    if interpolate:
        monoData = interpolateRT(rt, intensity)
    else:
        monoData = pd.DataFrame(dict(rt=rt, int=intensity))

    monoData = monoData.dropna(axis="index")
    if len(monoData) == 0:
        return MonoFit(0.0, 0.0, 0.0, 1 << FAIL_E_CURVE)

    i = monoData["int"].argmax()
    pos = monoData.iloc[i]

    # returns a smaller sigma than R
    def fit() -> tuple[np.ndarray, np.ndarray]:
        popt, pcov, *_ = curve_fit(
            normalIntensityFunction,
            monoData["rt"].to_numpy(),
            monoData["int"].to_numpy(),
            p0=(max(pos["rt"], 0.0), 1.0, max(pos["int"], 1e-5), 0.0),
            # method="lm"
            # maxfev=100,
            # mu,sigma, scale, baseline
            bounds=(
                np.array([0.0, 1e-6, 1e-6, 0.0]),
                np.array([np.inf, np.inf, np.inf, np.inf]),
            ),
        )
        return popt, pcov

    try:
        (popt, pcov), fail = catch_warnings(
            fit,
            peptide,
            "fitMonoEnvelope::curve_fit[%s]: %s",
        )
        fail = fail << FAIL_W_CURVE

    except (RuntimeError, TypeError) as e:
        if LOG_ERRORS:
            logger.error("fitMonoEnvelope::curve_fit[%s]: %s", peptide, e)
        return MonoFit(0.0, 0.0, 0.0, 1 << FAIL_E_CURVE)
    popt[1] = fabs(popt[1])  # ensure sigma is positive

    mu, sigma, kscale, base = popt

    lBoundary: float = max(mu - NSIGMA * sigma, rt[0])
    rBoundary: float = min(mu + NSIGMA * sigma, rt[-1])

    eicIds = (rt >= lBoundary) & (rt <= rBoundary)

    if eicIds.sum() == 0:
        return MonoFit(0.0, lBoundary, rBoundary, fail | 1 << FAIL_E_BOUNDARY, popt)

    if use_quadrature:

        def integrate() -> float:
            monoPeakArea, _ = quad(
                lambda rtv: np.interp(rtv, rt, intensity),
                lBoundary,
                rBoundary,
                limit=QUAD_LIMIT,
                # epsabs=1e-6,
                epsrel=1e-6,
            )
            return monoPeakArea

        monoPeakArea, fail2 = catch_warnings(
            integrate,
            peptide,
            "fitIsotopeEnvelopes::quad[%s]: %s",
        )
        fail |= fail2 << FAIL_W_QUAD

    else:
        monoPeakArea = base * (rBoundary - lBoundary) + kscale * sigma * RPI2 * (
            erf((rBoundary - mu) / sigma / R2) - erf((lBoundary - mu) / sigma / R2)
        )
    return MonoFit(
        monoPeakArea,
        lBoundary,
        rBoundary,
        fail,
        popt,
        pcov,
    )


def fitAllIntensities(im: np.ndarray, intensities: np.ndarray) -> np.ndarray:
    isotopeRegressions = np.array(
        list(fitIntensities(im, intensities)),
        dtype=np.float32,
    )
    return isotopeRegressions


def fitIsotopeEnvelopes(
    rawEIC: np.ndarray,  # [maxIso+2,len(rt),2]
    peptide: str = "unk peptide",
) -> tuple[IsotopeEnvelope | None, int]:
    from .config import WITH_ORIGIN, INTERPOLATE_INTENSITY

    monoEIC: np.ndarray = rawEIC[1]
    rt, intensity = monoEIC[:, 0], monoEIC[:, 1]

    monofit = fitMonoEnvelope(
        rt,
        intensity,
        interpolate=INTERPOLATE_INTENSITY,
        use_quadrature=USE_QUAD,
        peptide=peptide,
    )
    if monofit.fail:
        return None, monofit.fail
    eicIds = (rt >= monofit.lBoundary) & (rt <= monofit.rBoundary)
    intensity = intensity[eicIds]
    if WITH_ORIGIN:
        im = np.c_[np.ones(len(intensity), dtype=np.float32), intensity]
    else:
        im = intensity.reshape(-1, 1)

    intensities = rawEIC[:, eicIds, 1]
    # float[N=maxIso+2,3] # columns: adj_rsquared, slope (beta), alpha
    isotopeRegressions = fitAllIntensities(im, intensities)

    alpha, slopes = isotopeRegressions[:, 2], isotopeRegressions[:, 1]
    anynan = np.isnan(slopes)
    nfailed = np.sum(anynan)
    if nfailed == len(slopes):
        return None, 1 << FAIL_E_EICS

    a = alpha * (monofit.rBoundary - monofit.lBoundary)
    isotopeEnvelopes = ensure_pos(a + monofit.area * slopes)

    maxPeakArea = isotopeEnvelopes[1:].max()
    mono = isotopeEnvelopes[1]
    inv_ratio = isotopeEnvelopes[0] / mono if mono > 0.0 else np.inf
    adjustedRsq = isotopeRegressions[:, 0].astype(np.float32)
    avg_rsq = np.average(adjustedRsq[2:])  # skip -1 and mono
    return (
        IsotopeEnvelope(
            adjustedRsq=adjustedRsq,
            isotopeEnvelopes=isotopeEnvelopes.astype(np.float32),
            # mu, sigma, scale, baseline
            monoFitParams=monofit.popt.astype(
                np.float32,
            ),  # for plotting the fitted gaussian
            inv_ratio=inv_ratio,
            monoPeakArea=monofit.area,
            maxPeakArea=maxPeakArea,
            sigma=monofit.sigma,
            avg_rsq=float(avg_rsq),
        ),
        monofit.fail,
    )


EPS = 1e-10


def labelledEnvelopeCalculation(
    peptide: str,
    maxIso: int,
    isotopeEnvelopes: np.ndarray,
    settings: PeptideSettings,
) -> tuple[LabelledEnvelope, int]:
    assert len(isotopeEnvelopes) == maxIso + 2
    isotopeEnvelopes = isotopeEnvelopes[1:]  # skip -1 mzDiff

    pepinfo = PeptideInfo(peptide, settings)
    algo = settings_to_algo(settings)
    enrichments, isotopeEnvelopeBasis = algo.make_envelope_array(
        pepinfo,
        maxIso,
    )

    labelledEnvelopes, nnls_residual = nnls(isotopeEnvelopeBasis, isotopeEnvelopes)

    # ord=1 so that we have labelledEnvelopes like probabilities
    totalNNLSWeight = norm(labelledEnvelopes, ord=1) + EPS
    totalIntensityWeight = norm(isotopeEnvelopes, ord=2) + EPS

    # this is the same as LPF (other authors had different definitions of RIA)
    # ord=1
    relativeIsotopeAbundance = 1.0 - labelledEnvelopes[0] / totalNNLSWeight

    labelEnrichment2 = (enrichments @ labelledEnvelopes) / totalNNLSWeight
    labelEnrichment2 = min(
        max(0, labelEnrichment2),
        settings.maximumLabelEnrichment,
    )

    # this is mystifing
    # theoreticalMaxEnrichment = elementCount * totalNNLSWeight
    # labelEnrichment = (
    #     (np.array(range(len(labelledEnvelope))) * labelledEnvelope).sum()
    #     / theoreticalMaxEnrichment
    #     * settings.maximumLabelEnrichment
    # )

    theoreticalDist = isotopeEnvelopeBasis @ labelledEnvelopes
    natural_dist = settings_to_algo(settings).natural_dist
    heavyDistribution = heavy_dist(pepinfo, isotopeEnvelopes, natural_dist)
    nmax = len(isotopeEnvelopes)
    # nmax = max(len(theoreticalDist), len(heavyDistribution), len(isotopeEnvelopes))
    theoreticalDist = resize(theoreticalDist, nmax)
    heavyDistribution = resize(heavyDistribution, nmax)
    # isotopeEnvelopes = resize(isotopeEnvelopes, nmax)

    def fit() -> float:
        res = pearsonr(heavyDistribution, theoreticalDist)
        if np.isnan(res.statistic):  # type: ignore
            return 0.0
        return res.statistic  # type: ignore

    def fit2() -> float:
        res = pearsonr(isotopeEnvelopes, theoreticalDist)
        if np.isnan(res.statistic):  # type: ignore
            return 0.0
        return res.statistic  # type: ignore

    heavyCor, fail = catch_warnings(
        fit,
        peptide,
        "labelledEnvelopeCalculation::pearsonr[%s]: %s",
    )
    fail = fail << FAIL_W_PEARSON
    heavyCor2, fail2 = catch_warnings(
        fit2,
        peptide,
        "labelledEnvelopeCalculation::pearsonr2[%s]: %s",
    )
    fail |= fail2 << FAIL_W_PEARSON2

    return (
        LabelledEnvelope(
            labelledEnvelopes=labelledEnvelopes.astype(np.float32),
            theoreticalDist=theoreticalDist.astype(np.float32),
            relativeIsotopeAbundance=relativeIsotopeAbundance,
            enrichment=labelEnrichment2,
            # labelEnrichment2=labelEnrichment2,
            heavyCor=heavyCor,
            heavyCor2=heavyCor2,
            nnls_residual=nnls_residual,
            totalNNLSWeight=totalNNLSWeight,  # type: ignore
            totalIntensityWeight=totalIntensityWeight,  # type: ignore
        ),
        fail,
    )


def fitEnvelope(
    peptide: str,
    rawEIC: np.ndarray,  # [N==pep.maxIso+2,len(rt),2]
    settings: PeptideSettings,
) -> tuple[Envelope | None, int]:
    """Fit envelopes to raw EIC matrix of [mzranges, rt, intensity]"""
    iso, faile = fitIsotopeEnvelopes(rawEIC, peptide)
    if iso is None:
        return None, faile
    maxIso = len(rawEIC) - 2
    assert maxIso == len(iso.isotopeEnvelopes) - 2
    # heavy, failh = fitHeavyEnvelope(pep.peptide, iso)
    labelled, faill = labelledEnvelopeCalculation(
        peptide,
        maxIso,
        iso.isotopeEnvelopes,
        # heavy,
        settings,
    )
    return Envelope.make_all(iso, labelled), faill | faile



from __future__ import annotations

from typing import Callable
from typing import TYPE_CHECKING

import numpy as np
from scipy.fft import fft
from scipy.fft import ifft

from .utils import ATOMICPROPERTIES
from .utils import NAMES
from .utils import PeptideInfo
from .utils import PeptideSettings

if TYPE_CHECKING:
    from .types.numpytypes import NP1DF64Array, NP2DF64Array, NP1DF32Array


def isotopic_distribution(
    pepinfo: PeptideInfo,
    abundance: float | None = None,
    abundanceCutoff: float = 1e-10,
    maxMass: int = 100,
) -> NP1DF64Array:
    # pragma: no cover
    """return abundance of "heavy" peptides"""
    if abundance is None:
        ab = pepinfo.naturalAtomicAbundances
    else:
        ab = pepinfo.labelledAtomicAbundances(abundance)

    abundances = {pepinfo.settings.labelledElement: ab}

    formula = pepinfo.formula

    maxElements = np.sum(formula > 0, dtype=int)

    A = np.zeros((maxElements, maxMass), np.float64)

    elements = []
    for i, e in enumerate(NAMES):
        n = formula[i]
        if n > 0:
            elements.append(n)
            if e in abundances:
                a = abundances[e]
            else:
                # use natural abundance
                a = ATOMICPROPERTIES[e]["abundance"]
            A[len(elements) - 1, 0 : len(a)] = a

    tA = fft(A)

    ptA = np.ones((maxMass,), dtype=np.complex128)

    for i in range(maxElements):
        ptA *= tA[i] ** elements[i]

    riptA = np.real(ifft(ptA))  # type: ignore
    mx = np.max(np.where(riptA > abundanceCutoff))
    riptA = riptA[0 : int(mx) + 1]
    return np.fmax(riptA, 0.0)
    # return np.where(riptA > 0.0, riptA, 0.0)


def isotopic_distributionR(
    pepinfo: PeptideInfo,
    abundance: float | None = None,
    abundanceCutoff: float = 1e-10,
    maxMass: int = 100,
) -> NP1DF64Array:
    # pragma: no cover
    """return abundance of "heavy" peptides"""
    if abundance is None:
        ab = pepinfo.naturalAtomicAbundances
    else:
        ab = pepinfo.labelledAtomicAbundances(abundance)

    abundances = {pepinfo.settings.labelledElement: ab}

    formula = pepinfo.formula

    maxElements = np.sum(formula > 0, dtype=int)

    A = np.zeros((maxElements, maxMass), np.complex64)

    elements = []
    for i, e in enumerate(NAMES):
        n = formula[i]
        if n > 0:
            elements.append(n)
            if e in abundances:
                a = abundances[e]
            else:
                # use natural abundance
                a = ATOMICPROPERTIES[e]["abundance"]
            A[len(elements) - 1, 0 : len(a)] = a

            A[len(elements) - 1, :] = fft(A[len(elements) - 1, :], maxMass)[:maxMass]  # type: ignore

    tA = A

    ptA = np.ones((maxMass,), dtype=np.complex128)

    for i in range(maxElements):
        ptA *= tA[i] ** elements[i]

    riptA = np.real(ifft(ptA))  # type: ignore
    mx = np.max(np.where(riptA > abundanceCutoff))
    riptA = riptA[0 : int(mx) + 1]
    return np.fmax(riptA, 0.0)


def mk_maxIso(settings: PeptideSettings) -> Callable[[str], np.int32]:
    # pragma: no cover
    from .config import ABUNDANCE_CUTOFF

    def maxIso(peptide: str) -> np.int32:
        pepinfo = PeptideInfo(peptide, settings)
        r = isotopic_distribution(
            pepinfo,
            settings.maximumLabelEnrichment,
            abundanceCutoff=ABUNDANCE_CUTOFF,
        )
        return np.int32(len(r) - 1)

    return maxIso


def make_envelope_array(
    pepinfo: PeptideInfo,
    maxIso: int,
) -> tuple[NP1DF64Array, NP2DF64Array]:
    # enrichments = pepinfo.getEnrichments(maxIso)
    enrichments = pepinfo.settings.getEnrichmentsN(pepinfo.elementCount)

    isotope_envelope_basis = np.zeros(
        shape=(maxIso + 1, len(enrichments)),
        dtype=np.float64,
    )
    for i, element_enrichment_level in enumerate(enrichments):
        d = isotopic_distribution(pepinfo, element_enrichment_level)
        max_el = min(len(d), maxIso + 1)
        isotope_envelope_basis[:max_el, i] = d[:max_el]
    return enrichments, isotope_envelope_basis


def natural_dist(
    pepinfo: PeptideInfo,
    monoScale: float,
) -> NP1DF32Array:
    isod = isotopic_distribution(pepinfo)
    denom = isod[0]
    denom = denom if denom > 0.0 else 1.0
    return (monoScale * isod / denom).astype(np.float32)



from __future__ import annotations

from typing import NamedTuple
from typing import TypedDict

import numpy as np

# generated by df2class


class MZMLRow(TypedDict):
    retention_time_sec: float
    scanindex: int
    mzmax: float
    mzmin: float
    imax: float
    # mzranges: np.ndarray


# generated by df2class


class EICSJob(NamedTuple):
    eics: np.ndarray  # item: float32
    adjustedRsq: np.ndarray  # item: float32
    isotopeEnvelopes: np.ndarray  # item: float32
    monoFitParams: np.ndarray  # item: float32
    heavyDistribution: np.ndarray  # item: float32
    heavyFitParams: np.ndarray  # item: float32
    labelledEnvelopes: np.ndarray  # item: float32
    relativeIsotopeAbundance: float
    enrichment: float
    # labelEnrichment2: float
    heavyCor: float
    heavyCor2: float
    fail: int
    peptide_index: int
    peptide: str
    eics_shape: np.ndarray  # item: int64



from __future__ import annotations

from pathlib import Path
from typing import TypedDict

import pandas as pd  # type: ignore
from sqlalchemy import Boolean
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import inspect
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.engine import Connectable
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

from .utils import check_missing_columns
from .utils import file2engine
from .utils import secondary
# from sqlalchemy import BINARY

# func.count()
# pylint: disable=not-callable


class Base(DeclarativeBase):
    pass


class Modification(TypedDict):
    mass: float
    massdiff: float | None
    position: int
    site: str | None


class Peptide(Base):
    __tablename__ = "peptides"
    peptideid: Mapped[int] = mapped_column(Integer, primary_key=True)
    peptide: Mapped[str] = mapped_column(Text, nullable=False)
    assumed_charge: Mapped[int | None] = mapped_column(Integer)
    modcol: Mapped[str | None] = mapped_column(Text)
    calc_neutral_pep_mass: Mapped[float | None] = mapped_column(Float)
    precursor_neutral_mass: Mapped[float | None] = mapped_column(Float)
    peptideprophet_probability: Mapped[float | None] = mapped_column(Float)
    mz: Mapped[float | None] = mapped_column(Float)
    observed_mz: Mapped[float | None] = mapped_column(Float)
    retention_time_sec: Mapped[float | None] = mapped_column(Float)
    num_missed_cleavages: Mapped[int | None] = mapped_column(Integer)
    is_decoy: Mapped[bool | None] = mapped_column(Boolean)
    agg_count: Mapped[int | None] = mapped_column(Integer)
    protein_descr: Mapped[list[str] | None] = mapped_column(JSON)
    modifications: Mapped[list[Modification] | None] = mapped_column(JSON)
    run: Mapped[str | None] = mapped_column(String(64))
    isLabelledRun: Mapped[bool | None] = mapped_column(Boolean)
    fdr: Mapped[float | None] = mapped_column(Float)
    # ----- computed -----
    labelledElementCount: Mapped[int | None] = mapped_column(Integer)
    maxIso: Mapped[int | None] = mapped_column(Integer)
    mzranges: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    eics: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    # eics: Mapped[bytes | None] = mapped_column(BINARY, deferred=True)
    adjustedRsq: Mapped[list[float] | None] = mapped_column(
        JSON,
        deferred=True,
    )  # list[float]
    isotopeEnvelopes: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    monoPeakArea: Mapped[float | None] = mapped_column(Float)
    maxPeakArea: Mapped[float | None] = mapped_column(Float)
    monoFitParams: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    inv_ratio: Mapped[float | None] = mapped_column(Float)
    labelledEnvelopes: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    relativeIsotopeAbundance: Mapped[float | None] = mapped_column(Float)
    enrichment: Mapped[float | None] = mapped_column(Float)
    heavyCor: Mapped[float | None] = mapped_column(Float)
    heavyCor2: Mapped[float | None] = mapped_column(Float)
    theoreticalDist: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    nnls_residual: Mapped[float | None] = mapped_column(Float)
    totalNNLSWeight: Mapped[float] = mapped_column(Float)
    totalIntensityWeight: Mapped[float] = mapped_column(Float)
    fail: Mapped[int] = mapped_column(Integer)
    eics_shape: Mapped[list[int] | None] = mapped_column(JSON, deferred=True)
    # list of protein names from ProtXML file (needed for website)
    protein_names: Mapped[list[str] | None] = mapped_column(JSON)

    iso_peaks_nr_20: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_40: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_60: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_80: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )

    proteins: Mapped[list[Protein]] = relationship(
        lambda: Protein,
        secondary="peptides_proteins",
        back_populates="peptides",
    )

    def __repr__(self) -> str:  # pragma: no cover
        return f"Peptide({self.peptide},{self.mz}, {self.assumed_charge})"


# if you change this then you must change
# the columns iso_peaks_nr_{N} above
## @export
MINRSQ100 = [20, 40, 60, 80]


class PI(TypedDict):
    protein_name: str


class PeptideData(TypedDict):
    calc_neutral_pep_mass: float
    charge: int
    exp_tot_instances: float
    fpkm_adjusted_probability: float
    group_weight: str  # str of float
    initial_probability: float
    is_contributing_evidence: bool
    is_nondegenerate_evidence: bool
    mod_aminoacid_mass: float | None
    mod_nterm_mass: float | None
    modified_peptide: str | None
    n_enzymatic_termini: int
    n_instances: int
    n_sibling_peptides: float
    n_sibling_peptides_bin: int
    nsp_adjusted_probability: float
    peptide_group_designator: str
    peptide_parent_protein: list[PI] | None
    peptide_sequence: str
    weight: float


class Protein(Base):
    __tablename__ = "proteins"

    proteinid: Mapped[int] = mapped_column(Integer, primary_key=True)
    probability: Mapped[int | None] = mapped_column(Float)
    group_number: Mapped[int | None] = mapped_column(Integer)
    protein_name: Mapped[str | None] = mapped_column(Text)
    n_indistinguishable_proteins: Mapped[int | None] = mapped_column(Integer)
    percent_coverage: Mapped[float | None] = mapped_column(Float)
    unique_stripped_peptides: Mapped[list[str] | None] = mapped_column(JSON)
    group_sibling_id: Mapped[str | None] = mapped_column(String(16))
    total_number_peptides: Mapped[int | None] = mapped_column(Integer)
    total_number_distinct_peptides: Mapped[int | None] = mapped_column(Integer)
    confidence: Mapped[float | None] = mapped_column(Float)
    peptide: Mapped[list[PeptideData] | None] = mapped_column(JSON)
    protein_description: Mapped[str | None] = mapped_column(Text)
    pct_spectrum_ids: Mapped[float | None] = mapped_column(Float)
    prot_length: Mapped[int | None] = mapped_column(Integer)
    indistinguishable_protein: Mapped[list[str] | None] = mapped_column(JSON)
    lpf_median: Mapped[float | None] = mapped_column(Float)
    lpf_std: Mapped[float | None] = mapped_column(Float)
    num_found_peptides: Mapped[int | None] = mapped_column(Integer)

    peptides: Mapped[list[Peptide]] = relationship(
        Peptide,
        secondary="peptides_proteins",
        back_populates="proteins",
    )

    def __repr__(self) -> str:  # pragma: no cover
        return f"Protein({self.protein_name},prob={self.probability},npep={self.unique_stripped_peptides})"


class PepProt(Base):
    __tablename__ = "peptides_proteins"
    peptideid: Mapped[int] = mapped_column(
        Integer,
        ForeignKey(Peptide.peptideid),
        nullable=False,
        primary_key=True,
    )
    proteinid: Mapped[int] = mapped_column(
        Integer,
        ForeignKey(Protein.proteinid),
        nullable=False,
        primary_key=True,
    )


class Version(Base):
    __tablename__ = "version"
    version: Mapped[int] = mapped_column(Integer, nullable=False, primary_key=True)


def aggregate_lpf(
    prot: pd.DataFrame,
    pep: pd.DataFrame,
    secondary: pd.DataFrame,
) -> pd.DataFrame:
    m = secondary.merge(pep, how="inner", left_on="peptideid", right_on="peptideid")
    a = (
        m.groupby("proteinid")
        .agg(
            lpf_median=pd.NamedAgg(column="relativeIsotopeAbundance", aggfunc="median"),
            lpf_std=pd.NamedAgg(column="relativeIsotopeAbundance", aggfunc="std"),
            num_found_peptides=pd.NamedAgg(column="peptideid", aggfunc="count"),
        )
        .reset_index()
    )
    a.fillna(0.0, inplace=True)

    return prot.merge(a, how="left", left_on="proteinid", right_on="proteinid")


def save_db(
    pep: pd.DataFrame,
    prot: pd.DataFrame,
    engine: Engine,
) -> tuple[int, int, int]:
    from .utils import dehydrate_peptides, dehydrate_prot
    from .iso_peaks import add_iso_peaks
    from ..config import SQLITE_VERSION

    pep = dehydrate_peptides(pep)
    prot = dehydrate_prot(prot)
    # pep["peptideid"] = pep.index + 1
    prot["proteinid"] = prot.index + 1
    sec, names = secondary(pep, prot)
    # add protein_names concatenation to peptide
    pep = pep.merge(names, how="left", left_on="peptideid", right_on="peptideid")
    # add iso_peaks_nrNN peaks
    pep = add_iso_peaks(pep)
    # don't save any proteins that don't exist
    prot = prot[prot["proteinid"].isin(sec["proteinid"])]
    prot = aggregate_lpf(prot, pep, sec)

    Base.metadata.create_all(bind=engine)
    with engine.begin() as conn:
        npep = save_df(pep, Peptide, conn)
        nprot = save_df(prot, Protein, conn)
        nsec = save_df(sec, PepProt, conn)
        version = pd.DataFrame({"version": [SQLITE_VERSION]})
        save_df(version, Version, conn)

    # populate_isoPeaksNrFixed(engine, force=True)
    return npep, nprot, nsec


def save_df(
    df: pd.DataFrame,
    Orm: type[Base],
    conn: Connectable,
    chunksize: int = 1000,
) -> int:
    i = inspect(Orm)
    dtypes = {c.name: c.type for c in i.columns}

    cols = [n for n in dtypes if n in df.columns]

    return df[cols].to_sql(
        i.local_table.name,  # type: ignore
        con=conn,
        # need this for some reason....
        dtype=dtypes,  # type: ignore
        chunksize=chunksize,
        if_exists="append",
        index=False,
    )


def check_missing_pep(pep: pd.DataFrame) -> set[str]:
    return check_missing_columns(pep, Peptide)


def check_missing_prot(prot: pd.DataFrame) -> set[str]:
    return check_missing_columns(prot, Protein)


def save_to_file(
    sqlfile: Path,
    pep: pd.DataFrame,
    prot: pd.DataFrame,
) -> tuple[int, int, int]:
    if sqlfile.exists():  # pragma: no cover
        sqlfile.unlink()

    engine = file2engine(sqlfile)
    try:
        return save_db(pep, prot, engine)
    finally:
        engine.dispose()



from __future__ import annotations

import math
import pickle
from contextlib import suppress
from dataclasses import asdict
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import Any
from typing import Literal
from typing import TYPE_CHECKING
from typing import TypedDict

import numpy as np
import pandas as pd
from typing_extensions import ReadOnly


if TYPE_CHECKING:
    from .types.pepxmltypes import PepXMLRunRow
    from types import ModuleType
    from .types.numpytypes import (
        NP2DF32Array,
        NP1DF64Array,
        NP1DI32Array,
        NP1DNArray,
        NP3DF32Array,
    )


Algos = ["Latest", "2023"]


## @export
@dataclass(kw_only=True)
class PeptideSettings:
    rtTolerance: float = 15.0  # seconds
    mzTolerance: float = 1e-5
    labelledIsotopeNumber: int = 15
    labelledElement: str = "N"
    maximumLabelEnrichment: float = 0.95
    retentionTimeCorrection: Literal["UseInSample", "SimpleMedian"] = "SimpleMedian"
    useObservedMz: bool = False
    minProbabilityCutoff: float = 0.8
    enrichmentColumns: int = 10
    pipelineAlgo: Literal["Latest", "2023"] = "Latest"

    def __post_init__(self) -> None:
        err = self.validate()
        if err is not None:
            raise ValueError(err)

    def validate(self) -> str | None:
        msgs = []
        ok = True
        if self.labelledElement not in ATOMICPROPERTIES:
            msgs.append(f"unknown element {self.labelledElement}")
            ok = False
        elif (
            self.labelledIsotopeNumber
            not in ATOMICPROPERTIES[self.labelledElement]["isotopeNr"]
        ):
            ok = False
            msgs.append(
                f"unknown isotopeNumber {self.labelledIsotopeNumber} for {self.labelledElement}",
            )
        if getNr(self.labelledElement) == self.labelledIsotopeNumber:
            msgs.append(
                f"labelledIsotopeNumber {self.labelledIsotopeNumber} for {self.labelledElement} cannot be the stable isotope!",
            )
        if self.mzTolerance <= 0:
            msgs.append("mzTolerance must be postive!")

        if self.rtTolerance <= 0:
            msgs.append("rtTolerance must be postive!")
        if self.retentionTimeCorrection not in ["UseInSample", "SimpleMedian"]:
            msgs.append(
                'retentionTimeCorrection should be "UseInSample" or "SimpleMedian"',
            )
        if self.pipelineAlgo not in Algos:
            msgs.append(
                "pipelineAlgo should be one of: {Algos}",
            )
        # enrichment columns <=0 means use elementCount for number of columns
        # see PeptideInfo.getEnrichments
        # if self.enrichmentColumns <= 0: #
        #     msgs.append("enrichmentColumns must be postive!")

        if self.maximumLabelEnrichment <= 0 or self.maximumLabelEnrichment > 1.0:
            msgs.append(
                f"maximumLabelEnrichment {self.maximumLabelEnrichment} must be between zero and 1!",
            )

        if self.minProbabilityCutoff <= 0 or self.minProbabilityCutoff > 1.0:
            msgs.append(
                f"minProbabilityCutoff {self.minProbabilityCutoff} must be between zero and 1!",
            )

        if ok:  # so self.abundance works
            if self.abundance >= self.maximumLabelEnrichment:
                msgs.append(
                    f'maximumLabelEnrichment "{self.maximumLabelEnrichment}"'
                    f' cannot be less than the environmental isotopic abundance "{self.abundance}"',
                )
        if not msgs:
            return None
        return ", ".join(msgs)

    def hash(self) -> str:
        from hashlib import md5

        m = md5()
        for k, v in asdict(self).items():
            if isinstance(v, float):
                v = signif(v, 6)
            m.update(f"{k}={v}".encode())
        return m.hexdigest()

    def __hash__(self) -> int:
        return int(self.hash(), 16)

    @property
    def labelledAtomicAbundancesAtMaxEnrich(
        self,
    ) -> NP1DF64Array:
        return labelledAtomicAbundancesAtMaxEnrich(self)

    @property
    def iabundance(self) -> int:
        return getIabundance(self)

    @property
    def abundance(self) -> float:
        return getAbundance(self)

    @property
    def naturalAtomicAbundances(self) -> NP1DF64Array:
        return naturalAtomicAbundances(self.labelledElement)

    def labelledAtomicAbundances(self, abundance: float) -> NP1DF64Array:
        return labelledAtomicAbundances(
            self.labelledIsotopeNumber,
            self.labelledElement,
            abundance,
        )

    def getEnrichmentsN(self, ncols: int) -> NP1DF64Array:
        return getEnrichmentsN(ncols, self)

    def isoMzDiff(self, assumed_charge: int) -> float:
        return isoMzDiff(assumed_charge, self)

    def getElementCount(self, formula: NP1DI32Array) -> int:
        return getElementCount(formula, self.labelledElement)

    def getElementCountFromPeptide(self, peptide: str) -> int:
        return self.getElementCount(peptideFormula(peptide))

    def eic_mzranges(self, pep: PepXMLRunRow) -> NP2DF32Array:
        return eic_mzranges(pep, self)


class PeptideInfo:
    def __init__(self, peptide: str, settings: PeptideSettings) -> None:
        self.peptide = peptide
        self.settings = settings
        self._formula: NP1DI32Array | None = None

    @property
    def formula(self) -> NP1DI32Array:
        if self._formula is None:
            self._formula = peptideFormula(self.peptide)
        return self._formula

    @property
    def elementCount(self) -> int:
        return self.settings.getElementCount(self.formula)

    def getEnrichments(self, maxIso: int | None = None) -> NP1DF64Array:
        """enrichmentColumns array of enrichments"""
        ncols = self.settings.enrichmentColumns
        if ncols <= 0:
            ncols = self.elementCount
        elif ncols == 1:
            if maxIso is None:
                raise ValueError("getEnrichments: can't determine maxIso!")
            ncols = maxIso + 1  # make it square....
        return self.settings.getEnrichmentsN(ncols)

    @property
    def naturalAtomicAbundances(self) -> NP1DF64Array:
        return self.settings.naturalAtomicAbundances

    def labelledAtomicAbundances(self, abundance: float) -> NP1DF64Array:
        return self.settings.labelledAtomicAbundances(abundance)

    @property
    def iabundance(self) -> int:
        return self.settings.iabundance

    @property
    def abundance(self) -> float:
        return self.settings.abundance


def NPA(list_of_ints: list[float]) -> NP1DF64Array:
    return np.array(list_of_ints, dtype=np.float64)


def NPI(list_of_ints: list[int]) -> NP1DI32Array:
    return np.array(list_of_ints, dtype=np.int32)


NAMES = np.array(["C", "H", "O", "N", "P", "S", "Se"])
NAMES_DICT = {n: i for i, n in enumerate(NAMES)}
WATER = NPI([0, 2, 1, 0, 0, 0, 0])
# "C", "H", "O" ,"N", "P", "S" , "Se"
# minus H20
AMINOACIDS = {
    "A": NPI([3, 5, 1, 1, 0, 0, 0]),  # Alanine, ala
    "C": NPI([3, 5, 1, 1, 0, 1, 0]),  # Cysteine, cys
    "D": NPI([4, 5, 3, 1, 0, 0, 0]),  # Aspartic Acid, asx
    "E": NPI([5, 7, 3, 1, 0, 0, 0]),  # Glutamic acid, glu
    "F": NPI([9, 9, 1, 1, 0, 0, 0]),  # Phenylalanine, phe
    "G": NPI([2, 3, 1, 1, 0, 0, 0]),  # Glycine, gly
    "H": NPI([6, 7, 1, 3, 0, 0, 0]),  # Histidine, his
    "I": NPI([6, 11, 1, 1, 0, 0, 0]),  # Isoleucine, ile
    "K": NPI([6, 12, 1, 2, 0, 0, 0]),  # Lysine, lys
    "L": NPI([6, 11, 1, 1, 0, 0, 0]),  # Leucine, leu
    "M": NPI([5, 9, 1, 1, 0, 1, 0]),  # Methionine, met
    "N": NPI([4, 6, 2, 2, 0, 0, 0]),  # Asparagine, asn
    "P": NPI([5, 7, 1, 1, 0, 0, 0]),  # Proline, pro
    "Q": NPI([5, 8, 2, 2, 0, 0, 0]),  # Glutamine, gln
    "R": NPI([6, 12, 1, 4, 0, 0, 0]),  # Arginine, arg
    "S": NPI([3, 5, 2, 1, 0, 0, 0]),  # Serine, ser
    "T": NPI([4, 7, 2, 1, 0, 0, 0]),  # Threonine, thr
    "V": NPI([5, 9, 1, 1, 0, 0, 0]),  # Valine, val
    "W": NPI([11, 10, 1, 2, 0, 0, 0]),  # Tryptophan, trp
    "Y": NPI([9, 9, 2, 1, 0, 0, 0]),  # Tyrosine, tyr
    "a": NPI([2, 2, 1, 0, 0, 0, 0]),
    "c": NPI([5, 8, 2, 2, 0, 1, 0]),
    "m": NPI([5, 9, 2, 1, 0, 1, 0]),
    "U": NPI([3, 5, 1, 1, 0, 0, 1]),
}


class AtomicProperties(TypedDict):
    isotopeNr: ReadOnly[NP1DI32Array]
    abundance: ReadOnly[NP1DF64Array]
    mass: ReadOnly[NP1DF64Array]


# ****WARNING!*** the zeroth value is assumed to be the "natural" isotope.
ATOMICPROPERTIES: dict[str, AtomicProperties] = dict(
    C=AtomicProperties(
        isotopeNr=NPI([12, 13]),
        mass=NPA([12.0, 13.0033548378]),
        abundance=NPA([0.9889, 0.0111]),
    ),
    H=AtomicProperties(
        isotopeNr=NPI([1, 2]),
        mass=NPA([1.0078250321, 2.0141017780]),
        abundance=NPA([0.9998, 0.0001]),
    ),
    O=AtomicProperties(
        isotopeNr=NPI([16, 17, 18]),
        mass=NPA([15.9949146, 16.9991312, 17.9991603]),
        abundance=NPA([0.9976, 0.0004, 0.0020]),
    ),
    N=AtomicProperties(
        isotopeNr=NPI([14, 15]),
        mass=NPA([14.0030740052, 15.0001088984]),
        abundance=NPA([0.99633, 0.00367]),
    ),
    # need entry for formula calculation...
    P=AtomicProperties(
        isotopeNr=NPI([31]),
        mass=NPA([30.97376163]),
        abundance=NPA([1.0]),
    ),
    S=AtomicProperties(
        isotopeNr=NPI([32, 33, 34, 36]),
        mass=NPA([31.97207070, 32.97145843, 33.96786665, 35.96708062]),
        abundance=NPA([0.9502, 0.0075, 0.0421, 0.0002]),
    ),
    Se=AtomicProperties(
        isotopeNr=NPI([80, 78, 76, 82, 77, 74]),
        mass=NPA(
            [79.9165213, 77.9173091, 75.9192136, 81.9166994, 76.919914, 73.9224764],
        ),
        abundance=NPA([0.4961, 0.2377, 0.0937, 0.0873, 0.0763, 0.0089]),
    ),
)


## @export
def get_isotope_numbers() -> dict[str, list[int]]:
    return {
        elem: [int(n) for n in d["isotopeNr"]] for elem, d in ATOMICPROPERTIES.items()
    }


## @export
def get_element_names() -> list[str]:
    return list(NAMES)


def getNr(element: str) -> int:
    """get isotope number for most abundant element"""
    if element not in ATOMICPROPERTIES:
        raise ValueError(f'unknown element: "{element}"')
    at = ATOMICPROPERTIES[element]
    a = at["abundance"]
    return int(at["isotopeNr"][a.argmax()])


def getDefaultIsotopeNr(element: str) -> int:
    """get default isotope number for experiment"""
    if element not in ATOMICPROPERTIES:
        raise ValueError(f'unknown element: "{element}"')
    at = ATOMICPROPERTIES[element]
    a = at["isotopeNr"]
    return int(a[1])


def getElementCount(formula: NP1DI32Array, element: str) -> int:
    return formula[NAMES == element][0]


def okNr(element: str) -> set[int]:
    """set of possible atomic numbers"""
    at = ATOMICPROPERTIES[element]
    a = at["isotopeNr"]
    return {int(i) for i in a}


# return C H O N P S Se count list
def peptideFormula(peptideSequence: str) -> NP1DI32Array:
    return reduce(
        lambda total, aa: total + AMINOACIDS[aa],
        list(peptideSequence),
        WATER,
    )


def getIabundance(settings: PeptideSettings) -> int:
    a = ATOMICPROPERTIES[settings.labelledElement]
    b = a["isotopeNr"] == settings.labelledIsotopeNumber
    return np.where(b)[0][0]


def getAbundance(setting: PeptideSettings) -> float:
    a = ATOMICPROPERTIES[setting.labelledElement]
    b = a["isotopeNr"] == setting.labelledIsotopeNumber
    return a["abundance"][b][0]


def getEnrichmentsN(
    elementCount: int,
    settings: PeptideSettings,
) -> NP1DF64Array:
    return np.linspace(
        settings.abundance,
        settings.maximumLabelEnrichment,
        elementCount,
        endpoint=True,
    )


def labelledAtomicAbundances(
    labelledIsotopeNumber: int,
    element: str,
    abundance: float,
) -> NP1DF64Array:
    if element in ATOMICPROPERTIES:
        if abundance < 0.0 or abundance > 1.0:
            raise ValueError(f'abundance "{abundance}" must be between 0 and 1')
        ap = ATOMICPROPERTIES[element]
        if labelledIsotopeNumber in ap["isotopeNr"]:
            a = ap["abundance"]
            label = ap["isotopeNr"] == labelledIsotopeNumber
            adjustedAbundance = (1 - abundance) * a / np.sum(a[~label])

            adjustedAbundance[label] = abundance
            return adjustedAbundance
    raise ValueError(f"unknown element: {element}[{labelledIsotopeNumber}]")


def naturalAtomicAbundances(
    element: str,
) -> NP1DF64Array:
    if element in ATOMICPROPERTIES:
        ap = ATOMICPROPERTIES[element]
        return ap["abundance"]

    raise ValueError(f'unknown element: "{element}"')


def labelledAtomicAbundancesAtMaxEnrich(
    settings: PeptideSettings,
) -> NP1DF64Array:
    return labelledAtomicAbundances(
        settings.labelledIsotopeNumber,
        settings.labelledElement,
        settings.maximumLabelEnrichment,
    )


# def okiso(settings: PeptideSettings) -> bool:
#     props = ATOMICPROPERTIES[settings.labelledElement]
#     isotopeNr, abundance = props["isotopeNr"], props["abundance"]
#     idx = (isotopeNr == settings.labelledIsotopeNumber)[0]
#     ab = (abundance == np.max(abundance))[0]
#     return idx != ab


def isoMzDiff(assumed_charge: int, settings: PeptideSettings) -> float:
    props = ATOMICPROPERTIES[settings.labelledElement]
    mass, abundance = props["mass"], props["abundance"]
    pos = props["isotopeNr"] == settings.labelledIsotopeNumber
    isoDiff = mass[pos][0] - mass[abundance.argmax()]

    isoMzDiff = isoDiff / assumed_charge
    return isoMzDiff


@dataclass(kw_only=True)
class MZEICRecord:
    assumed_charge: int
    mz: float
    observed_mz: float
    maxIso: int

    @staticmethod
    def from_series(s: pd.Series):
        return MZEICRecord(**s.to_dict())


def eic_mzranges(
    pep: PepXMLRunRow | MZEICRecord,
    settings: PeptideSettings,
) -> NP2DF32Array:
    # requires
    # maxIso, mz, assumed_charge and
    # (settings.{labelledElement,labelledIsotopeNumber}, mzTolerance)
    isoDiff = settings.isoMzDiff(pep.assumed_charge)
    mz = pep.observed_mz if settings.useObservedMz else pep.mz
    tol = mz * settings.mzTolerance
    # NOTE: pep.maxIso might be a floating point number if
    # pep is actually a pd.Series
    isoRange = range(-1, round(pep.maxIso) + 1)
    mzranges = []
    for iso in isoRange:
        v = mz + iso * isoDiff
        mzranges.append([v - tol, v + tol])

    return np.array(mzranges, dtype=np.float32)


# @overload
# def resize(a: NP1DF64Array, width: int) -> NP1DF64Array: ...


# @overload
# def resize(a: NP1DF32Array, width: int) -> NP1DF32Array: ...


# @overload
# def resize(a: NP1DI32Array, width: int) -> NP1DI32Array: ...


def resize(a: NP1DNArray, width: int) -> NP1DNArray:
    d = width - len(a)
    if d > 0:
        return np.pad(a, (0, d), "constant", constant_values=(0, 0))
    if d < 0:
        return a[:width]
    return a


def ensure_pos(
    a: NP1DNArray,
) -> NP1DNArray:
    # use np.fmax if we want to set a[i] to zero if a[i] is NaN
    return np.maximum(a, 0.0)
    # return np.where(a >= 0, a, 0.0)


# def roundit(n: float, nsig: int = 3) -> int:
#     from math import log10

#     assert n > 0, str(n)
#     pwr = round(log10(n))
#     pwr = max(0, pwr - nsig)
#     num = 10**pwr
#     return round(n / num) * num


def signif(x: float | int, digits: int = 6) -> float | int:
    if x == 0 or not math.isfinite(x):
        return x
    digits -= math.ceil(math.log10(abs(x)))
    return round(x, digits)


def human(num: int, suffix: str = "B", scale: int = 1) -> str:
    if not num:
        return f"0{suffix}"
    num *= scale
    magnitude = int(math.floor(math.log(abs(num), 1000)))
    val = num / math.pow(1000, magnitude)
    if magnitude > 7:
        return "{:.1f}{}{}".format(val, "Y", suffix)
    return "{:3.1f}{}{}".format(
        val,
        ["", "k", "M", "G", "T", "P", "E", "Z"][magnitude],
        suffix,
    )


def rmfiles(files: list[Path]) -> None:
    for f in files:
        with suppress(OSError):
            Path(f).unlink(missing_ok=True)


def duplicate_rows(df: pd.DataFrame, on: list[str]) -> pd.DataFrame:
    return df[df.duplicated(on, keep=False)]


def fdr(df: pd.DataFrame) -> float:  # pragma: no cover
    decoy = df["is_decoy"].sum()
    denom = len(df) - decoy
    return float(decoy) / denom if denom > 0 else np.inf


def calculate_fdrs(
    df: pd.DataFrame,
    nbins: int = 50,
    score: str = "peptideprophet_probability",
) -> pd.Series:
    """Requires peptideprophet_probability, is_decoy"""
    # df['fdr'] = calculate_fdrs(df)
    from math import ceil

    decoys = df["is_decoy"]
    pbins = df[score].apply(lambda p: ceil(p * nbins))

    mx = int(pbins.max())

    fdrs: dict[int, float] = {}
    for binno in range(mx + 1):
        target = pbins >= binno
        n = target.sum()
        if n == 0:  # pragma: no cover
            fdrs[binno] = fdrs.get(binno - 1, 0.0)
            continue
        ndecoys = decoys[target].sum()
        denom = n - ndecoys
        fdrs[binno] = ndecoys / denom if denom > 0 else np.inf
    return pbins.apply(lambda binno: fdrs[binno]).astype(np.float32)


class Apply:
    @staticmethod
    def eics_reshape(
        s: pd.Series,
    ) -> NP3DF32Array:  # pragma: no cover
        return s["eics"].reshape(s["eics_shape"]).astype(np.float32)

    @staticmethod
    def eics_reshape_np(
        s: pd.Series,
    ) -> NP3DF32Array:
        return np.array(s["eics"]).reshape(s["eics_shape"]).astype(np.float32)

    @staticmethod
    def eics_reshape_bytes(
        s: pd.Series,
    ) -> NP3DF32Array:
        return pickle.loads(s["eics"]).reshape(s["eics_shape"]).astype(np.float32)

    @staticmethod
    def modcol(mods: list[dict[str, int | float | str]], n: int = 3) -> str:
        mods = sorted(mods, key=lambda d: int(d["position"]))
        return ":".join([f"{float(d['mass']):.{n}f}@{d['position']}" for d in mods])

    @staticmethod
    def protxml_modcol(peptided: dict[str, Any], n: int = 3) -> pd.Series:
        name = peptided["peptide_sequence"]
        if "mod_aminoacid_mass" not in peptided:
            return pd.Series(dict(peptide=name, modcol=""))
        return pd.Series(
            dict(peptide=name, modcol=Apply.modcol(peptided["mod_aminoacid_mass"], n)),
        )

    # @staticmethod
    # def binrt(rt: float, rttol: float = 15.0) -> int: # pragma: no cover
    #     return round(rt / rttol)

    # @staticmethod
    # def binmz(mz: float, mztol: float = 10e-6) -> int:  # pragma: no cover
    #     return round(mz / mztol)


class IO:
    def __init__(self, filename: Path | str, df: pd.DataFrame | None = None):
        self.df = df
        self.filename = Path(filename)

    def save_df(self, index: bool = False) -> None:
        assert self.df is not None
        try:
            if not self.filename.parent.exists():  # pragma: no cover
                self.filename.parent.mkdir(parents=True, exist_ok=True)
            self.df.to_parquet(self.filename, index=index)

        except (Exception, KeyboardInterrupt):  # pragma: no cover
            rmfiles([self.filename])  # cleanup
            raise

    def read_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        return pd.read_parquet(self.filename, columns=columns)  # , use_threads=True)


def df_can_write(fmt: Literal["excel", "parquet"]) -> bool:
    from io import BytesIO

    df = pd.DataFrame({"x": [0]})
    try:
        func = getattr(df, f"to_{fmt}", None)
        if func is None:  # pragma: no cover
            return False
        func(BytesIO())
        return True
    except ModuleNotFoundError:  # pragma: no cover
        return False


def df_can_write_parquet() -> bool:
    return df_can_write("parquet")


def df_can_write_excel() -> bool:
    return df_can_write("excel")


def getsize(fname: Path) -> int:
    return fname.stat().st_size


# @overload
# def array_split(df: pd.Series, num: int) -> list[pd.Series]: ...


# @overload
# def array_split(df: pd.DataFrame, num: int) -> list[pd.DataFrame]: ...


def array_split(
    df: pd.DataFrame,
    num: int,
) -> list[pd.DataFrame]:
    w, e = divmod(len(df), num)
    if e > 0:
        w += 1
    ret = []
    for i in range(0, len(df), w):
        ret.append(df.iloc[i : i + w])
    return ret


def has_package(package: str) -> bool:
    return get_package(package) is not None


def get_package(package: str) -> ModuleType | None:
    import importlib

    try:
        return importlib.import_module(package)

    except ModuleNotFoundError:  # pragma: no cover
        return None


def compress(sqlite: Path, remove: bool = False) -> Path:
    import shutil
    import gzip

    ret = sqlite.with_suffix(sqlite.suffix + ".gz")
    with sqlite.open("rb") as f_in:
        with gzip.open(ret, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    assert ret.exists()
    if remove:
        sqlite.unlink(missing_ok=True)
    return ret


def decompress(gzsqlite: Path) -> Path:
    import shutil
    import gzip

    ret = gzsqlite.with_suffix("")
    with gzip.open(gzsqlite, "rb") as f_in:
        with ret.open("wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    assert ret.exists()
    ret.touch()  # update mtime
    return ret


## @export
def decompressto(gzsqlite: Path, outsqlite: Path | None = None) -> Path:
    import shutil
    import gzip

    if outsqlite is None:
        outsqlite = gzsqlite.with_suffix("")
    with gzip.open(gzsqlite, "rb") as f_in:
        with outsqlite.open("wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    assert outsqlite.exists()
    outsqlite.touch()  # update mtime
    return outsqlite



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from ..logger import logger
from ..resourcefiles import MzMLResourceFile
from ..utils import IO
from .cmd import Command


@dataclass
class DinoRunner:
    jarpath: Path
    javapath: Path = Path("java")

    def cmdline(self) -> list[str]:
        from shutil import which

        javapath = which(self.javapath or "java")
        if javapath is None:
            return []

        return [str(javapath), "-jar", str(self.jarpath.absolute())]

    @classmethod
    def from_config(cls) -> DinoRunner | None:
        from ..config import DINOSAUR_JAR, JAVA_PATH

        if not DINOSAUR_JAR:
            return None
        if JAVA_PATH is None:
            return None
        return DinoRunner(jarpath=Path(DINOSAUR_JAR), javapath=Path(JAVA_PATH))

    def can_run(self) -> bool:
        from shutil import which

        return self.jarpath.exists() and which(self.javapath) is not None

    def run(self, mzml: MzMLResourceFile) -> None | int:
        cmdline = self.cmdline() + [
            "--verbose",
            f"--outDir={mzml.cache_dir or '.'}",
            f"--outName={mzml.hash}",
            str(mzml.original.absolute()),
        ]

        cmd = Command(cmdline)

        for line in cmd.output:
            logger.info("dinosaur: %s", line)
        if cmd.returncode:
            logger.error("failed to generate dinosaur for %s", mzml.original.name)
            return cmd.returncode

        output = mzml.cache_dir / f"{mzml.hash}.features.tsv"

        df = pd.read_csv(output, sep="\t")
        output.unlink()
        output = mzml.cache_dinosaur()
        logger.info("writing dinosaur: %s", output)
        IO(output, df).save_df()
        return cmd.returncode


def mzml_dinosaur(mzml: MzMLResourceFile) -> None:
    dino = DinoRunner.from_config()
    if dino is None or not dino.can_run():
        return

    dino.run(mzml)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.fft import irfft
from scipy.fft import irfftn
from scipy.fft import next_fast_len
from scipy.fft import rfft
from scipy.fft import rfftn

from .utils import PeptideInfo

if TYPE_CHECKING:
    from typing import Callable
    from .utils import PeptideSettings
    from .types.numpytypes import NP1DF64Array, NP2DF64Array, NP1DF32Array


def nextfast(natoms: int) -> int:
    return next_fast_len(natoms + 1, real=True)  # type: ignore


# see https://pubs.acs.org/doi/pdf/10.1021/ac500108n
def fractions1D(abundances: NP1DF64Array, natoms: int) -> NP1DF64Array:
    nlen = nextfast(natoms)
    c = np.zeros(nlen, np.float64)
    c[:2] = abundances
    return irfft(
        rfft(c, overwrite_x=True) ** natoms,  # type: ignore
        nlen,
        overwrite_x=True,
    )[: natoms + 1]  # type: ignore


def fractionsND(
    abundances: NP1DF64Array,
    natoms: int,
) -> np.ndarray[tuple[int, ...], np.dtype[np.float64]]:
    nlen = nextfast(natoms)
    shape = tuple([nlen] * (len(abundances) - 1))
    m = len(shape)
    c = np.zeros(shape, np.float64)

    z = lambda: [0] * m
    # abundance zero is the main stable isotope
    c[tuple(z())] = abundances[0]

    for i, a in enumerate(abundances[1:]):
        idx = z()
        idx[i] = 1
        c[tuple(idx)] = a

    return irfftn(rfftn(c, overwrite_x=True) ** natoms, c.shape, overwrite_x=True)  # type: ignore


def fractions(abundances: NP1DF64Array, iabundance: int, natoms: int) -> NP1DF64Array:
    if len(abundances) == 2:
        return fractions1D(abundances, natoms)
    r = fractionsND(abundances, natoms)
    # pull out iabundance axis
    idx = [0] * r.ndim
    idx[iabundance - 1] = slice(natoms + 1)  # type: ignore
    return r[tuple(idx)]


## ---- API -----


def isotopic_distribution(
    pepinfo: PeptideInfo,
    abundance: float | None = None,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]:
    """implements paper https://pubs.acs.org/doi/pdf/10.1021/ac500108n"""
    # also https://dx.doi.org/10.1021/ac500108n

    if abundance is None:
        abundances = pepinfo.naturalAtomicAbundances
    else:
        abundances = pepinfo.labelledAtomicAbundances(abundance)

    r = fractions(abundances, pepinfo.iabundance, pepinfo.elementCount)
    return np.fmax(r, 0.0)


def mk_maxIso(settings: PeptideSettings) -> Callable[[str], np.int32]:
    from .config import ABUNDANCE_CUTOFF

    abundances = settings.labelledAtomicAbundancesAtMaxEnrich
    iabundance = settings.iabundance

    def maxIso(peptide: str) -> np.int32:
        elementCount = settings.getElementCountFromPeptide(peptide)
        r = fractions(abundances, iabundance, elementCount)
        mx = np.max(np.where(r > ABUNDANCE_CUTOFF))
        mx = min(mx + 1, elementCount)
        return np.int32(mx)

    return maxIso


def make_envelope_array(
    pepinfo: PeptideInfo,
    maxIso: int,
) -> tuple[NP1DF64Array, NP2DF64Array]:
    enrichments = pepinfo.getEnrichments(maxIso)

    isotope_envelope_basis = np.zeros(
        shape=(maxIso + 1, len(enrichments)),
        dtype=np.float64,
    )
    for i, element_enrichment_level in enumerate(enrichments):
        d = isotopic_distribution(pepinfo, element_enrichment_level)
        max_el = min(len(d), maxIso + 1)
        isotope_envelope_basis[:max_el, i] = d[:max_el]
    return enrichments, isotope_envelope_basis


def natural_dist(
    pepinfo: PeptideInfo,
    monoScale: float,
) -> NP1DF32Array:
    isod = isotopic_distribution(pepinfo)
    denom = isod[0]
    denom = denom if denom > 0.0 else 1.0
    return (monoScale * isod / denom).astype(np.float32)



from __future__ import annotations

import math
from typing import Any
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import func
from sqlalchemy import inspect
from sqlalchemy import select
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy.engine import Engine

if TYPE_CHECKING:
    from sqlalchemy import Select, ColumnElement
    from pathlib import Path


## @export
def file2engine(filename: str | Path, *, echo: bool = False) -> Engine:
    return create_engine(f"sqlite:///{filename}", future=True, echo=echo)


def dataframe_sql(df: pd.DataFrame) -> str:  # pragma: no cover
    res = []
    for cname, dtype in df.dtypes.to_dict().items():
        k = dtype.kind
        if k == "f":
            a = "Float"
        elif k == "i":
            a = "Integer"
        elif k == "b":
            a = "Boolean"
        else:
            typ = type(df[cname].iloc[0])
            if typ is str:
                m = df[cname].apply(len).max()
                if m > 4096:
                    a = "Text"
                else:
                    a = f"String({m})"
            else:
                a = "JSON"
        r = f"    {cname} = Column({a})"
        res.append(r)
    return "\n".join(res)


def check_missing_columns(df: pd.DataFrame, Orm: type) -> set[str]:
    i: Table = inspect(Orm)
    dtypes = {c.name for c in i.columns}
    pepcols = {str(s) for s in df.columns}
    return dtypes - pepcols


def dehydrate_floatarray(arr: np.ndarray) -> list[float]:
    return arr.astype(np.float32).tolist()


def dehydrate_intarray(arr: np.ndarray) -> list[int]:
    return [int(f) for f in arr]


# def rehydrate_pep(df: pd.DataFrame) -> pd.DataFrame:
#     def sto(strarr):
#         return np.array(strarr, dtype=object)

#     df["modifications"] = df["modifications"].apply(sto)  # type: ignore
#     return df


def dehydrate_peptide_col(
    peptide_list: list[dict[str, Any]] | np.ndarray,
) -> list[dict[str, Any] | list[Any]]:
    def fix(v: dict[str, Any]) -> dict[str, Any] | list[Any]:
        if isinstance(v, np.ndarray):
            return dehydrate_peptide_col(v)
        if isinstance(v, dict):
            return {k: fix(vv) for k, vv in v.items()}
        return v

    return [fix(value) for value in peptide_list]


def dehydrate_prot(df: pd.DataFrame) -> pd.DataFrame:
    df["peptide"] = df.peptide.apply(dehydrate_peptide_col)
    df["indistinguishable_protein"] = df.indistinguishable_protein.apply(list)
    df["unique_stripped_peptides"] = df.unique_stripped_peptides.apply(list)
    return df


def secondary(
    pep: pd.DataFrame,
    prot: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    exploded = prot[["proteinid", "protein_name", "unique_stripped_peptides"]].explode(
        "unique_stripped_peptides",
    )
    dfm = pep[["peptideid", "peptide"]].merge(
        exploded,
        how="inner",
        left_on="peptide",
        right_on="unique_stripped_peptides",
    )
    sec = dfm[["peptideid", "proteinid"]]
    names = (
        dfm[["peptideid", "protein_name"]]
        .groupby("peptideid")
        .agg({"protein_name": lambda s: sorted(set(s))})
    )
    names = names.reset_index().rename(columns=dict(protein_name="protein_names"))
    return sec, names


def eics_to_bytes(a: np.ndarray) -> bytes:
    return a.astype(np.float32).dumps()


def dehydrate_peptides(df: pd.DataFrame) -> pd.DataFrame:
    if "result_index" in df.columns:
        df.rename(columns=dict(result_index="peptideid"), inplace=True)
    for col in ["proteins", "protein_descr", "modifications"]:
        if col in df.columns:
            df[col] = df[col].apply(list)
    for col in [
        "mzranges",
        "eics",
        "adjustedRsq",
        "isotopeEnvelopes",
        "monoFitParams",
        "heavyDistribution",
        "labelledEnvelopes",
        "heavyFitParams",
        "theoreticalDist",
    ]:
        if col in df.columns:
            df[col] = df[col].apply(dehydrate_floatarray)
    # for col in ["eics"]:
    #     if col in df.columns:
    #         df[col] = df[col].apply(eics_to_bytes)
    for col in ["eics_shape"]:
        if col in df.columns:
            df[col] = df[col].apply(dehydrate_intarray)
    return df


def rehydrate_peptides(turn: pd.DataFrame) -> pd.DataFrame:
    from ..utils import Apply

    def reshape2(v):
        return np.array(v).reshape((-1, 2))

    func = Apply.eics_reshape_np
    # func = Apply.eics_reshape_bytes

    if "eics" in turn.columns and "eics_shape" in turn.columns:
        turn["eics"] = turn[["eics", "eics_shape"]].apply(
            func,
            axis=1,
        )
        turn.drop(columns=["eics_shape"], inplace=True)
    if "mzranges" in turn.columns:
        turn["mzranges"] = turn["mzranges"].apply(reshape2)  # type: ignore

    return turn


def show_table_create(name: str, engine: Engine) -> str:
    with engine.connect() as conn:
        return str(
            conn.execute(
                text(f"SELECT sql FROM sqlite_schema WHERE name = '{name}'"),
            ).scalar(),
        )


def round10(x: float | int) -> float | int:
    if x == 0 or not math.isfinite(x):
        return x
    digits = math.ceil(math.log10(abs(x)))
    sign = 1 if x > 0 else -1
    return sign * 10**digits


def precision(
    number: float | int,
    significant_figures: int = 3,
    *,
    shift: int = 0,
) -> float:
    if number == 0:
        return 0

    # Calculate the order of magnitude
    order_of_magnitude = math.floor(math.log10(abs(number)))

    # Calculate the scaling factor to shift the number so the first significant digit is in the units place
    scaling_factor = 10 ** (significant_figures - 1 - order_of_magnitude)

    # Scale the number, round it, and then scale it back
    nn = number * scaling_factor
    nn += shift
    return round(nn) / scaling_factor


def std(engine: Engine, col: Column) -> float:
    q = select(
        func.avg(col * col),
        func.avg(col),
        func.count(col),  # pylint: disable=not-callable
    )
    # q = q.select_from(col.table)
    with engine.connect() as conn:
        xx, x, c = conn.execute(q).one()
        r = c / (c - 1) if c > 1 else 1
        v = (xx - x * x) * r
        return math.sqrt(v)


def sql_to_str(q: Select[Any] | ColumnElement[Any]) -> str:
    from sqlalchemy.dialects import sqlite

    return str(
        q.compile(dialect=sqlite.dialect(), compile_kwargs={"literal_binds": True}),
    )



from __future__ import annotations

import hashlib
from abc import ABC
from abc import abstractmethod
from pathlib import Path

from typing_extensions import override

from .exts import DINOSAUR
from .exts import EICS
from .exts import EXT
from .exts import MZMAP
from .exts import MZML
from .exts import PEPXML
from .exts import PROTXML
from .exts import RESULT_EXT


def hash256(filename: Path, bufsize: int = 4096 * 8, algo: str = "sha256") -> str:
    sha256_hash = hashlib.new(algo, usedforsecurity=False)
    with filename.open("rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(bufsize), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def fileok(cached: list[Path], original: Path) -> bool:
    if not cached:  # pragma: no cover
        return False
    if not original.is_file():
        return False
    omtime = original.stat().st_mtime
    for cache in cached:
        if not cache.is_file() or omtime > cache.stat().st_mtime:
            return False
    return True


class BaseResourceFile(ABC):
    def __init__(
        self,
        original: Path | str,
        cache_dir: Path | str | None = None,
    ):
        self.original = Path(original).expanduser().absolute()
        if cache_dir is None:
            self.cache_dir = self.original.parent
        else:
            self.cache_dir = Path(cache_dir).expanduser().absolute()
        self._hash: str | None = None

    @property
    def name(self) -> str:
        return self.original.name

    def exists(self) -> bool:
        return self.original.exists()

    def extok_(self, *ext: str) -> bool:
        targets = [self.cache_file_(e) for e in ext]
        return fileok(targets, self.original)

    def cache_file_(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.hash + ext)

    def cache_ok(self) -> bool:
        ok = self.all_cache_files()
        return fileok(ok, self.original)

    def ensure_cache_dir(self):
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    @property
    def hash(self) -> str:
        if self._hash is not None:
            return self._hash

        self._hash = self.get_hash_()
        return self._hash

    def get_hash_(self) -> str:
        return hash256(self.original)

    def cleanup(self) -> None:
        for pth in self.all_cache_files():
            try:
                pth.unlink(missing_ok=True)
            except OSError:  # pragma: no cover
                pass

    @abstractmethod
    def all_cache_files(self) -> list[Path]:  # pragma: no cover
        return []

    def get_lock_files_(self) -> list[Path]:
        return [path.parent / (path.name + ".lock") for path in self.all_cache_files()]

    def lock(self) -> None:
        """Don't delete a cache file "X" if it has an "X.lock" file"""
        for path in self.get_lock_files_():
            path.touch()

    def unlock(self) -> None:
        for path in self.get_lock_files_():
            try:
                path.unlink(missing_ok=True)
            except OSError:  # pragma: no cover
                pass

    def is_locked(self) -> bool:
        for path in self.get_lock_files_():
            if path.exists():
                return True
        return False


class MzMLResourceFile(BaseResourceFile):
    @property
    def runName(self) -> str:
        """Name possibly in pep.xml file as <spectrum_query spectrum="{runName}.{start_scan}.{end_scan}.0 ...>"""
        return self.original.stem

    def cache_mzml(self) -> Path:
        return self.cache_file_(MZML)

    def cache_memmap(self) -> Path:
        return self.cache_file_(MZMAP)

    def cache_dinosaur(self) -> Path:  # pragma: no cover
        return self.cache_file_(DINOSAUR)

    def cache_mzml_ok(self) -> bool:
        return self.extok_(MZML)

    def cache_memmap_ok(self) -> bool:
        return self.extok_(MZMAP)

    # def cache_dinosaur_ok(self):
    #     from .dinosaur.dinosaur import DinoRunner

    #     ok = self.extok_(DINOSAUR)
    #     if ok:
    #         return True

    #     dino = DinoRunner.from_config()
    #     # no dino file check if we can generate it...
    #     if dino is not None and dino.can_run():
    #         return False
    #     # ok we can't generate it so cache is "fine..."
    #     return True

    @override
    def all_cache_files(self) -> list[Path]:
        return [self.cache_mzml(), self.cache_memmap()]

    # @override
    # def cache_ok(self) -> bool:
    #     ok = super().cache_ok()
    #     if not ok:
    #         return ok
    #     return self.cache_dinosaur_ok()


class MzMLResourceFileLocal(MzMLResourceFile):
    def __init__(
        self,
        original: Path | str,
        cache_dir: Path | str | None = None,
        *,
        name: str | Path | None = None,
        ext: str = EXT,
    ):
        super().__init__(original, cache_dir)
        if name is None:
            name = self.name
        self.outname = Path(name)
        if not ext.startswith("."):
            ext = "." + ext
        self.ext = ext

    @override
    def cache_file_(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.outname.with_suffix(ext))

    @override
    def cache_mzml(self) -> Path:
        return self.cache_file_(self.ext)

    @override
    def cache_memmap(self) -> Path:
        return self.cache_file_(".mzi")

    @override
    def cache_mzml_ok(self) -> bool:
        return self.extok_(self.ext)

    @override
    def cache_memmap_ok(self) -> bool:
        return self.extok_(".mzi")


class PepXMLResourceFile(BaseResourceFile):
    def cache_pepxml(self) -> Path:
        return self.cache_file_(PEPXML)

    def cache_pepxml_ok(self) -> bool:
        return self.extok_(PEPXML)

    @override
    def all_cache_files(self) -> list[Path]:
        return [self.cache_pepxml()]


class PepXMLResourceFileLocal(PepXMLResourceFile):
    def __init__(
        self,
        original: Path | str,
        cache_dir: Path | str | None = None,
        *,
        name: str | Path | None = None,
        ext: str = EXT,
    ):
        super().__init__(original, cache_dir)
        if name is None:
            name = self.name
        self.outname = Path(name)
        if not ext.startswith("."):
            ext = "." + ext
        self.ext = ext

    @override
    def cache_file_(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.outname.with_suffix(ext))

    @override
    def cache_pepxml(self) -> Path:
        return self.cache_file_(self.ext)

    @override
    def cache_pepxml_ok(self) -> bool:
        return self.extok_(self.ext)


class ProtXMLResourceFile(BaseResourceFile):
    def cache_protxml(self) -> Path:
        return self.cache_file_(PROTXML)

    def cache_protxml_ok(self) -> bool:
        return self.extok_(PROTXML)

    @override
    def all_cache_files(self) -> list[Path]:
        return [self.cache_protxml()]


class ProtXMLResourceFileLocal(ProtXMLResourceFile):
    def __init__(
        self,
        original: Path | str,
        cache_dir: Path | str | None = None,
        *,
        name: str | Path | None = None,
        ext: str = EXT,
    ):
        super().__init__(original, cache_dir)
        if name is None:
            name = self.name
        self.outname = Path(name)
        if not ext.startswith("."):
            ext = "." + ext
        self.ext = ext

    @override
    def cache_protxml(self) -> Path:
        return self.cache_file_(self.ext)

    @override
    def cache_protxml_ok(self) -> bool:
        return self.extok_(self.ext)

    @override
    def cache_file_(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.outname.with_suffix(ext))


class ResultsResourceFile(BaseResourceFile):
    def result_file(self) -> Path:
        return self.cache_result()

    def cache_result(self) -> Path:
        return self.cache_file_(RESULT_EXT)

    def cache_pepxml(self) -> Path:
        return self.cache_file_(PEPXML)

    def cache_envelope(self) -> Path:
        return self.cache_file_(EICS)

    def cache_partial_envelope(self, idx: int) -> Path:
        return self.cache_file_(f"-{idx}{EICS}")

    def has_result(self) -> bool:
        return self.result_file().exists()

    def has_compressed_result(self) -> bool:
        return self.result_file().with_suffix(RESULT_EXT + ".gz").exists()

    @override
    def cache_file_(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.original.name + ext)

    @override
    def all_cache_files(self) -> list[Path]:
        return [self.cache_result(), self.cache_pepxml(), self.cache_envelope()]


class ResourceFiles:
    def __init__(
        self,
        pepxmls: list[PepXMLResourceFile],
        protxml: ProtXMLResourceFile,
        mzmlfiles: list[MzMLResourceFile],
    ):
        self.pepxmls = pepxmls
        self.protxml = protxml
        self.mzmlfiles = mzmlfiles
        self.cache_dirs = {
            *[s.cache_dir for s in pepxmls],
            *[s.cache_dir for s in mzmlfiles],
        }

    def all_cache_files_(self) -> list[BaseResourceFile]:
        ret = [s for s in [*self.pepxmls, *self.mzmlfiles] if s.exists()]
        if self.protxml.exists():
            ret.append(self.protxml)
        return ret

    def lock(self) -> None:
        for resourcefile in self.all_cache_files_():
            resourcefile.lock()

    def unlock(self) -> None:
        for resourcefile in self.all_cache_files_():
            resourcefile.unlock()

    def is_locked(self) -> bool:
        for resourcefile in self.all_cache_files_():
            if resourcefile.is_locked():
                return True
        return False

    def todo(self) -> list[BaseResourceFile]:
        ret: list[BaseResourceFile] = []
        for pepxml in self.pepxmls:
            if pepxml.exists() and not pepxml.cache_pepxml_ok():
                ret.append(pepxml)
        for mzml in self.mzmlfiles:
            if mzml.exists() and not mzml.cache_mzml_ok():
                ret.append(mzml)
        if self.protxml.exists() and not self.protxml.cache_ok():
            ret.append(self.protxml)

        return ret

    def ensure_directories(self) -> None:
        for cd in self.cache_dirs:
            if not cd.is_dir():  # pragma: no cover
                cd.mkdir(parents=True, exist_ok=True)



from __future__ import annotations


MIN_RT = 2
ABUNDANCE_CUTOFF = 0.01
# split jobs into this many spectra
# set to zero to not split
NSPECTRA = 10000
# quadrature limit for estimating area under curve (unused now)
# see USE_QUAD in fitenvelopes.py
# QUAD_LIMIT = 500
# https://docs.python.org/3/library/logging.html#logrecord-attributes
LOG_FORMAT = "%(levelname)s|[%(asctime)s]|%(process)d| %(message)s"

# XCMS_STEP = 0.0
PEPXML_CHUNKS = 1000

MAIL_SUBJECT = "turnover pipeline"
# default "from" sender
MAIL_SENDER = "turnover-pipeline@uwa.edu.au"
# set to None or 'none' to stop any emailing
# e.g. MAIL_SERVER="uwa-edu-au.mail.protection.outlook.com"
MAIL_SERVER = "mailhost"
MAIL_TIMEOUT = 20.0
# e.g. "https://protein-turnover.plantenergy.org/inspect/{jobid}"
INSPECT_URL: str | None = None
# email template with {job:TurnoverJob, url:str}
MAIL_TEXT = """Protein Turnover Job "{job.job_name}" has <b>finished</b>!{url}."""
# set to True for production version (hides debug click commands)

INTERPOLATE_INTENSITY = True

# add origin...
WITH_ORIGIN = False

# Dinosaur
JAVA_PATH = None  # "java"
DINOSAUR_JAR = None

SQLITE_VERSION = 1

COMPRESS_SLEEP = 60 * 60  # wait an hour between scans for uncompressed .sqlite files
COMPRESS_AGE = 60 * 60 * 6  # remove sqlite files older than 6 hours



from __future__ import annotations

from itertools import islice
from pathlib import Path
from typing import Any
from typing import Iterator

import pandas as pd
from more_itertools import ichunked
from pymzml.run import Reader
from typing_extensions import override

from .logger import log_iterator
from .pepxml_reader import chop_spectrum

# https://github.com/levitsky/pyteomics/blob/b9fabac991d2da524b4700af871e56f6a255f924/pyteomics/pepxml.py#L349
# taken from pyteomics.pepxml:DataFrame

# UNUSED: see pepxml_reader.py
if False:  #  pragma: no cover
    from pyteomics.pepxml import PepXML

    def fix_mods(mods: list[dict]) -> Iterator[dict]:
        for m in mods:
            if "variable" in m:
                yield dict(
                    mass=m["mass"],
                    position=m["position"],
                    massdiff=m["variable"],
                )

            elif "static" in m:
                yield dict(
                    mass=m["mass"],
                    position=m["position"],
                    massdiff=m["static"],
                )
            else:
                # both these are optional but one must exist?
                yield dict(mass=m["mass"], position=m["position"])

    def dict_to_series(item: dict[str, Any]) -> dict[str, Any]:
        info: dict[str, Any] = {}
        for k, v in item.items():
            if isinstance(v, (str, int, float)):
                info[k] = v
        if "search_hit" in item:
            sh = item["search_hit"][0]  # first element
            proteins = sh.pop("proteins")

            # list of dicts to dict of lists
            prot_dict: dict[str, list[Any]] = {}
            for p in proteins:
                for k in p:
                    prot_dict[k] = []
            for p in proteins:
                for k, v in prot_dict.items():
                    v.append(p.get(k))
            prot_dict["proteins"] = prot_dict.pop("protein")
            prot_dict["protein_descr"] = [
                p if p is not None else "" for p in prot_dict["protein_descr"]
            ]
            info.update(prot_dict)

            info.update(sh.pop("search_score"))
            mods = sh.pop("modifications", [])
            info["modifications"] = list(fix_mods(mods))

            for k, v in sh.items():
                if isinstance(v, (str, int, float)):
                    info[k] = v
            if "analysis_result" in sh:
                for ar in sh["analysis_result"]:
                    if ar["analysis"] == "peptideprophet":
                        result = ar["peptideprophet_result"]
                        try:
                            info.update(result["parameter"])
                        except KeyError:
                            pass
                        info["peptideprophet_probability"] = result["probability"]
                        info["peptideprophet_ntt_prob"] = result["all_ntt_prob"]
                    elif ar["analysis"] == "interprophet":
                        result = ar["interprophet_result"]
                        info.update(result["parameter"])
                        info["interprophet_probability"] = result["probability"]
                        info["interprophet_ntt_prob"] = result["all_ntt_prob"]

        info["run"] = chop_spectrum(info["spectrum"])
        # is_rejected: Potential use in future for user manual validation (0 or 1)
        if "is_rejected" in info:
            info.pop("is_rejected")
        if "xcorr" in info:
            info["searchEngineScore"] = info["xcorr"]  # COMET
        elif "ionscore" in info:
            info["searchEngineScore"] = info["ionscore"]  # MASCOT

        return info

    def pepxml_chunk(
        chunk: Iterator[dict[str, Any]],
        **kwargs: Any,
    ) -> pd.DataFrame | None:
        lst = [dict_to_series(r) for r in chunk]
        if not lst:
            return None
        return pd.DataFrame(lst, **kwargs)

    def PepXMLDataFrame(
        filename: Path,
        stop: int | None = None,
        start: int = 0,
        step: int = 1,
        level: int = 0,
        number_of_bg_processes: int = 0,
        **kwargs: Any,
    ) -> pd.DataFrame:
        from .config import PEPXML_CHUNKS

        pep = Path(filename)

        it = PepXML(str(filename), use_index=True)  # need to get total
        total = len(it)
        if stop is not None or start != 0 or step != 1:
            # find size
            if stop is None:
                stop = total
            else:
                stop = min(stop, total)
            r = range(start, stop, step)
            total = len(r)
            it = islice(it, r.start, r.stop, r.step)
        ret = []
        it = log_iterator(
            it,
            total=total,
            level=level,
            desc=f"{pep.name}",
            number_of_bg_processes=number_of_bg_processes,
        )
        for chunk in ichunked(it, PEPXML_CHUNKS):
            df = pepxml_chunk(chunk, **kwargs)
            if df is None:
                continue
            ret.append(df)

        if not ret:
            raise RuntimeError(f"{pep.name}: no runs!")

        return pd.concat(ret, axis=0, ignore_index=True)


# at least version 2.5.2
class PyMzMLReader(Reader):
    @override
    def term_is_a_member(self, tested_term, member_of_term):
        """shutup to the KeyError print statement"""
        try:
            self.OT[tested_term]
        except KeyError:  # pragma: no cover
            return False
        return super().term_is_a_member(tested_term, member_of_term)


def ProtXmlDataFrame(
    protxml: Path,
    level: int = 0,
    number_of_bg_processes: int = 0,
) -> pd.DataFrame:
    from pyteomics.protxml import ProtXML
    from .logger import log_iterator

    prot = ProtXML(str(protxml))
    total = len(prot)

    def gen_items(prot: ProtXML) -> Iterator[dict[str, Any]]:
        with prot as f:
            for item in f:
                info = {}
                # probability 1.0
                # group_number 1
                # protein [{}, {}]
                for k, v in item.items():
                    if isinstance(v, (str, int, float)):
                        info[k] = v
                if "protein" in item:
                    for prot in item["protein"]:
                        # protein_name contam_sp|P02769|ALBU_BOVIN
                        # n_indistinguishable_proteins 1
                        # probability 1.0
                        # percent_coverage 53.0
                        # unique_stripped_peptides ['AEFVEVTK', 'ATEEQLK', 'CCTESLVNR', 'CCTKPESER', 'DDPHACYSTVFDK', 'EACFAVEGPK', ]
                        # group_sibling_id a
                        # total_number_peptides 72
                        # total_number_distinct_peptides 72
                        # confidence 1.0
                        # protein_description Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4
                        # pct_spectrum_ids 0.322
                        # prot_length 607
                        # peptide [{}, {}]

                        #
                        # peptide : {
                        # 'peptide_sequence': 'CCTESLVNR',
                        # 'charge': 2,
                        # 'initial_probability': 0.999,
                        # 'nsp_adjusted_probability': 0.9996,
                        # 'fpkm_adjusted_probability': 0.9996,
                        # 'weight': 1.0,
                        # 'group_weight': '1.00',
                        # 'n_enzymatic_termini': 2,
                        # 'n_sibling_peptides': 41.14,
                        # 'n_sibling_peptides_bin': 9,
                        # 'n_instances': 4,
                        # 'exp_tot_instances': 3.93,
                        # 'calc_neutral_pep_mass': 1137.4906,
                        # 'peptide_parent_protein': [{'protein_name': 'sp|P02768|ALBU_HUMAN'}],
                        # 'is_nondegenerate_evidence': False,
                        # 'is_contributing_evidence': True,
                        # 'modified_peptide': 'CCTESLVNR',
                        # 'mod_aminoacid_mass': [{position:1, mass:160.03}, ...]
                        # }
                        out = dict(info)
                        out.update(prot)
                        if "indistinguishable_protein" in out:
                            out["indistinguishable_protein"] = [
                                p["protein_name"]
                                for p in out["indistinguishable_protein"]
                            ]
                        else:
                            out["indistinguishable_protein"] = []
                        yield out

    it = log_iterator(
        gen_items(prot),
        total=total,
        level=level,
        desc=f"{protxml.name}",
        number_of_bg_processes=number_of_bg_processes,
    )

    return pd.DataFrame(it)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Literal

import click

from .cli import cli
from .cli import CONFIG
from .cli import Config
from .cli import pass_config


def web_options(f: Callable) -> Callable:
    f = click.option(
        "-s",
        "--server",
        type=click.Choice(["gunicorn", "waitress", "flask"]),
        help="which wsgi server to run",
    )(f)
    f = click.option(
        "-n",
        "--no-browse",
        is_flag=True,
        help="don't open web application in browser",
    )(f)
    f = click.option(
        "--port",
        default=8000,
        help="port to run webservice on",
        show_default=True,
    )(f)

    return f


WebServer = Literal["flask", "gunicorn", "waitress"]


@cli.command(
    epilog="""

Running a webserver + turnover backend.

If not specifed with options or in a configuration file.
job results will be stored in `~/turnover_jobs` (created if not existing)
and file caches will use `~/turnover_cache` (created in not existing).

\b
Configuration
=============

If a configuration file is specified then you can specifiy
JOBSDIR, CACHEDIR.

The program will look for a file `~/.turnover.toml` for website
configuration if no `--web-config` is given.

Run `turnover web` and browse to "Configuration" for more information.

You can give arguments to the underlying webserver with arguments after
a `--`, e.g.:

`turnover web --server=gunicorn --workers=4 -- --workers=2 --access-logfile=-`

will tell gunicorn to fire up 2 workers and use 4 background workers for turnover itself.
Also show access logs on stdout.
""",
)
@click.option(
    "-w",
    "--workers",
    type=int,
    help="number of background workers [default: half the number of cpus]",
)
@web_options
@click.option(
    "-c",
    "--web-config",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="configuration file for web browser",
)
@click.option(
    "--jobs-dir",
    type=click.Path(file_okay=False),
    help="directory to run job [default: directory of ~/turnover_jobs]",
)
@click.option(
    "--cache-dir",
    type=click.Path(file_okay=False),
    help="directory to store file caches [default: directory of ~/turnover_cache]",
)
@click.option(
    "-v",
    "--view-only",
    is_flag=True,
    help="only view data",
)
@click.option(
    "--mailhost",
    help="specify MAILHOST to allow background jobs to email user when done. Default is to not send emails",
    default="none",
)
@click.option(
    "--compress-result",
    is_flag=True,
    help="compress the resultant sqlite file to save space (use with care as it slows things down)",
)
@click.argument("server_options", nargs=-1)
@pass_config
def web(
    cfg: Config,
    no_browse: bool,
    workers: int,
    web_config: str | None,
    server: WebServer | None = None,
    jobs_dir: str | None = None,
    cache_dir: str | None = None,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
    view_only: bool = False,
    mailhost: str | None = None,
    compress_result: bool = False,
) -> None:  # pragma: no cover
    """Run full website (requires protein-turnover-website)."""
    from os import cpu_count
    from .web import WebRunner

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)
    defaults: dict[str, Any] = {}
    if cache_dir is not None:
        defaults["CACHEDIR"] = cache_dir

    if jobs_dir is not None:
        defaults["JOBSDIR"] = [jobs_dir]
    wr = WebRunner(
        browse=not no_browse,
        workers=workers,
        web_config=web_config,
        server=server,
        configfile=cfg.user_config,
        defaults=defaults,
        port=port,
        server_options=server_options,
        view_only=view_only,
        mailhost=mailhost,
        compress_result=compress_result,
    )

    wr.run()


@cli.command()
@web_options
@click.argument("jobfile", type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.argument("server_options", nargs=-1)
@pass_config
def view(
    cfg: Config,
    no_browse: bool,
    jobfile: str,
    server: WebServer | None = None,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
) -> None:  # pragma: no cover
    """View a completed run in a browser (requires protein-turnover-website)"""
    from pathlib import Path
    from .web import WebRunner
    from .jobs import TurnoverJob
    from .resourcefiles import ResultsResourceFile

    jf = Path(jobfile).expanduser().absolute()
    if not jf.exists():
        click.secho(f"no such file {jobfile}", fg="red")
        raise click.Abort()

    try:
        job = TurnoverJob.restore(jf)
    except Exception as e:
        click.secho(
            f'Can\'t read "{jf.name}" as a TOML file: {e}',
            fg="red",
            bold=True,
            err=True,
        )
        raise click.Abort()
    jobsddir = jf.parent
    results = ResultsResourceFile(job.jobid, jobsddir)
    iscompressed = results.has_compressed_result()
    if not iscompressed and not results.has_result():
        raise ValueError(f"can't find output data file {results.result_file()}")
    dataid = jf.stem
    defaults = {
        "JOBSDIR": [str(jobsddir)],
        "DATAID": dataid,
        "COMPRESS_RESULT": iscompressed,
    }

    wr = WebRunner(
        browse=not no_browse,
        workers=1,
        web_config=None,
        server=server,
        configfile=cfg.user_config,
        defaults=defaults,
        port=port,
        view_only=True,
        wsgi_app="protein_turnover_website.wsgi_view",
        server_options=server_options,
        mailhost="none",
        compress_result=iscompressed,
    )

    wr.run()


@cli.command(hidden=True)
@web_options
@click.argument("server_options", nargs=-1)
@pass_config
def config(
    cfg: Config,
    no_browse: bool,
    server: WebServer | None = None,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
) -> None:
    """View Configuration information"""
    from .web import WebRunner

    wr = WebRunner(
        browse=not no_browse,
        workers=1,
        web_config=None,
        defaults={"DATAID": "noid"},
        server=server,
        port=port,
        # extra=web_options,
        view_only=True,
        wsgi_app="protein_turnover_website.wsgi_view",
        page="configuration.html",
        server_options=server_options,
        mailhost="none",
    )
    wr.run()


@cli.command()
@click.option(
    "-o",
    "--out",
    type=click.Path(dir_okay=False),
    help="write configuration file here [default ~/.turnover.toml]",
)
@click.option("--force", is_flag=True, help="force overwriting of any existing file")
@click.option("--without-website", is_flag=True, help="don't add website configuration")
def init_config(force: bool, without_website: bool, out: str | None) -> None:
    """Create a default configuration file"""
    from pathlib import Path

    from .web import has_package, dump_config

    if out is None:
        configfile = CONFIG
    else:
        configfile = Path(out)

    if configfile.exists() and not force:
        click.secho(f"{configfile} file exists! Use --force to overwrite.")
        return
    if not without_website and not has_package("protein_turnover_website"):
        click.secho(
            "Please install protein_turnover_website for website config [pip install protein-turnover-website]!",
            fg="red",
            err=True,
        )
        raise click.Abort()

    if not has_package("pyarrow"):
        click.secho(
            "Please install pyarrow!",
            fg="red",
            err=True,
        )
        raise click.Abort()

    dump_config(configfile, not without_website)
